# ----------------------------------------------------------------------------
# Copyright (c) 2018-2023, QIIME 2 development team.
#
# Distributed under the terms of the Modified BSD License.
#
# The full license is in the file LICENSE, distributed with this software.
# ----------------------------------------------------------------------------
import os
from subprocess import CalledProcessError

import numpy as np
import numpy.testing as npt
import biom
import skbio
import pkg_resources

from qiime2.plugin.testing import TestPluginBase
from qiime2 import Artifact

from ..beta import bray_curtis, jaccard, METRICS


# ----------------------------Non-Phylogenetic---------------------------------
class BrayCurtisTests(TestPluginBase):
    package = 'q2_diversity_lib.tests'

    def setUp(self):
        super().setUp()
        self.input_table = biom.Table(np.array([[1, 0, 999, 1],
                                                [0, .5, 0, 1],
                                                [0, 0, 1, 1]]),
                                      ['A', 'B', 'C'],
                                      ['S1', 'S2', 'S3', 'S4'])

        self.expected = skbio.DistanceMatrix(
                [[0.0000000, 1, 0.998001998, 0.5],
                 [1, 0.0000000, 1, .714285714],
                 [0.998001998, 1, 0.0000000, .996011964],
                 [0.5, .714285714, .996011964, 0.0000000]],
                ids=['S1', 'S2', 'S3', 'S4'])

    def test_method(self):
        actual = bray_curtis(table=self.input_table, n_jobs=1)
        self.assertEqual(actual.ids, self.expected.ids)
        for id1 in actual.ids:
            for id2 in actual.ids:
                npt.assert_almost_equal(actual[id1, id2],
                                        self.expected[id1, id2])

    def test_accepted_types_have_consistent_behavior_rarefied(self):
        rarefied_table = biom.Table(np.array([[3, 1, 2, 6],
                                             [1, 3, 2, 0],
                                             [2, 2, 2, 0]]),
                                    ['A', 'B', 'C'],
                                    ['S1', 'S2', 'S3', 'S4'])

        normalized_rarefied_table = rarefied_table.norm(axis='sample',
                                                        inplace=False)
        self.expected = skbio.DistanceMatrix(
                [[0.0000000, 0.333333333333333, 0.1666666667, 0.5],
                 [0.333333333333333, 0.0000000, 0.1666666667, 0.833333333333],
                 [0.1666666667, 0.1666666667, 0.0000000, 0.66666666667],
                 [0.5, 0.833333333333, 0.66666666667, 0.0000000]],
                ids=['S1', 'S2', 'S3', 'S4'])

        accepted_tables = [rarefied_table, normalized_rarefied_table]
        for table in accepted_tables:
            actual = bray_curtis(table=table)
            self.assertEqual(actual.ids, self.expected.ids)
            for id1 in actual.ids:
                for id2 in actual.ids:
                    npt.assert_almost_equal(actual[id1, id2],
                                            self.expected[id1, id2])

    def test_accepted_types_have_inconsistent_behavior_unrarefied(self):
        freq_table = self.input_table
        relative_freq_table = self.input_table.norm(axis='sample',
                                                    inplace=False)
        # count data should pass, as in test_method
        actual = bray_curtis(table=freq_table)
        for id1 in actual.ids:
            for id2 in actual.ids:
                npt.assert_almost_equal(actual[id1, id2],
                                        self.expected[id1, id2])

        # normalized data should fail
        actual = bray_curtis(table=relative_freq_table)
        with self.assertRaisesRegex(AssertionError,
                                    "not almost equal"):
            for id1 in actual.ids:
                for id2 in actual.ids:
                    npt.assert_almost_equal(actual[id1, id2],
                                            self.expected[id1, id2])

    def test_bray_curtis_relative_frequency(self):
        input_table = biom.Table(
            np.array([
                [0.3, 0, 0.77, 0.5],
                [0.1, 0, 0.15, 0.25],
                [0.6, 1, 0.08, 0.25]
            ]),
            ['A', 'B', 'C'],
            ['S1', 'S2', 'S3', 'S4']
        )
        expected = skbio.DistanceMatrix(
            [
                [0.0000000, 0.4, 0.52, 0.35],
                [0.4, 0.0000000, 0.92, 0.75],
                [0.52, 0.92, 0.0000000, 0.27],
                [0.35, 0.75, 0.27, 0.0000000]
            ],
            ids=['S1', 'S2', 'S3', 'S4']
        )
        actual = bray_curtis(table=input_table, n_jobs=1)
        self.assertEqual(actual.ids, self.expected.ids)
        for id1 in actual.ids:
            for id2 in actual.ids:
                npt.assert_almost_equal(
                    actual[id1, id2], expected[id1, id2]
                )


class JaccardTests(TestPluginBase):
    package = 'q2_diversity_lib.tests'

    def setUp(self):
        super().setUp()
        self.freq_table = biom.Table(np.array([[1, 0, 999, 1],
                                               [0, .5, 0, 1],
                                               [0, 0, 1, 1]]),
                                     ['A', 'B', 'C'],
                                     ['S1', 'S2', 'S3', 'S4'])
        self.expected = skbio.DistanceMatrix(
                [[0.0000000, 1, 0.5, 0.6666667],
                 [1, 0.0000000, 1, 0.6666667],
                 [0.5, 1, 0.0000000, 0.3333333],
                 [0.6666667, 0.6666667, 0.3333333, 0.0000000]],
                ids=['S1', 'S2', 'S3', 'S4'])

        self.relative_freq_table = self.freq_table.norm(axis='sample',
                                                        inplace=False)
        self.p_a_table = self.freq_table.pa(inplace=False)

    def test_method(self):
        actual = jaccard(table=self.freq_table, n_jobs=1)
        self.assertEqual(actual.ids, self.expected.ids)
        for id1 in actual.ids:
            for id2 in actual.ids:
                npt.assert_almost_equal(actual[id1, id2],
                                        self.expected[id1, id2])

    def test_accepted_types_have_consistent_behavior(self):
        accepted_tables = [self.freq_table, self.relative_freq_table,
                           self.p_a_table]
        for table in accepted_tables:
            actual = jaccard(table=table)
            self.assertEqual(actual.ids, self.expected.ids)
            for id1 in actual.ids:
                for id2 in actual.ids:
                    npt.assert_almost_equal(actual[id1, id2],
                                            self.expected[id1, id2])


# ---------------------------------Phylogenetic-------------------------------
class UnweightedUnifrac(TestPluginBase):
    package = 'q2_diversity_lib.tests'

    def artifact(self, semantic_type, fp):
        return Artifact.import_data(semantic_type, self.get_data_path(fp))

    def setUp(self):
        super().setUp()

        self.fn = self.plugin.actions['unweighted_unifrac']

        # expected computed with skbio.diversity.beta_diversity
        self.expected = skbio.DistanceMatrix([[0.00, 0.25, 0.25],
                                             [0.25, 0.00, 0.00],
                                             [0.25, 0.00, 0.00]],
                                             ids=['S1', 'S2', 'S3'])

        self.tbl = self.artifact('FeatureTable[Frequency]',
                                 'two_feature_table.biom')
        self.tre = self.artifact('Phylogeny[Rooted]', 'three_feature.tree')

    def test_method(self):
        actual_art, = self.fn(self.tbl, self.tre)
        actual = actual_art.view(skbio.DistanceMatrix)
        self.assertEqual(actual.ids, self.expected.ids)
        for id1 in actual.ids:
            for id2 in actual.ids:
                npt.assert_almost_equal(actual[id1, id2],
                                        self.expected[id1, id2])

    def test_method_bypass_tips(self):
        actual_art, = self.fn(self.tbl, self.tre, bypass_tips=True)
        actual = actual_art.view(skbio.DistanceMatrix)
        self.assertEqual(actual.ids, self.expected.ids)

    def test_accepted_types_have_consistent_behavior(self):
        rf_tbl = self.artifact('FeatureTable[RelativeFrequency]',
                               'two_feature_rf_table.biom')
        pa_tbl = self.artifact('FeatureTable[PresenceAbsence]',
                               'two_feature_p_a_table.biom')

        for table in [self.tbl, rf_tbl, pa_tbl]:
            actual_art, = self.fn(table=table, phylogeny=self.tre)
            actual = actual_art.view(skbio.DistanceMatrix)
            self.assertEqual(actual.ids, self.expected.ids)
            for id1 in actual.ids:
                for id2 in actual.ids:
                    npt.assert_almost_equal(actual[id1, id2],
                                            self.expected[id1, id2])

    def test_missing_tips_tree(self):
        tre = self.artifact('Phylogeny[Rooted]', 'root_only.tree')
        with self.assertRaises(CalledProcessError):
            obs = self.fn(self.tbl, tre)
            self.assertTrue('not a subset of the tree tips', obs.stderr)

    def test_extra_tip_tree(self):
        tbl = self.artifact('FeatureTable[Frequency]',
                            'two_feature_table.biom')

        tre_2 = self.artifact('Phylogeny[Rooted]', 'two_feature.tree')
        obs_2_art, = self.fn(tbl, tre_2)
        obs_2 = obs_2_art.view(skbio.DistanceMatrix)

        tre_3 = self.artifact('Phylogeny[Rooted]', 'three_feature.tree')
        obs_3_art, = self.fn(tbl, tre_3)
        obs_3 = obs_3_art.view(skbio.DistanceMatrix)

        self.assertEqual(obs_2, obs_3)


class WeightedUnifrac(TestPluginBase):
    package = 'q2_diversity_lib.tests'

    def artifact(self, semantic_type, fp):
        return Artifact.import_data(semantic_type, self.get_data_path(fp))

    def setUp(self):
        super().setUp()

        self.fn = self.plugin.actions['weighted_unifrac']

        # expected computed with diversity.beta_phylogenetic (weighted_unifrac)
        self.expected = skbio.DistanceMatrix(
            np.array([0.44656238, 0.23771096, 0.30489123, 0.23446002,
                      0.65723575, 0.44911772, 0.381904, 0.69144829,
                      0.39611776, 0.36568012, 0.53377975, 0.48908025,
                      0.35155196, 0.28318669, 0.57376916, 0.23395746,
                      0.24658122, 0.60271637, 0.39802552, 0.36567394,
                      0.68062701, 0.36862049, 0.48350632, 0.33024631,
                      0.33266697, 0.53464744, 0.74605075, 0.53951035,
                      0.49680733, 0.79178838, 0.37109012, 0.52629343,
                      0.22118218, 0.32400805, 0.43189708, 0.59705893]),
            ids=('10084.PC.481', '10084.PC.593', '10084.PC.356',
                 '10084.PC.355', '10084.PC.354', '10084.PC.636',
                 '10084.PC.635', '10084.PC.607', '10084.PC.634'))

        self.tbl = self.artifact('FeatureTable[Frequency]', 'crawford.biom')
        self.tre = self.artifact('Phylogeny[Rooted]', 'crawford.nwk')

    def test_method(self):
        actual_art, = self.fn(self.tbl, self.tre)
        actual = actual_art.view(skbio.DistanceMatrix)
        self.assertEqual(actual.ids, self.expected.ids)
        for id1 in actual.ids:
            for id2 in actual.ids:
                npt.assert_almost_equal(actual[id1, id2],
                                        self.expected[id1, id2])

    def test_method_bypass_tips(self):
        actual_art, = self.fn(self.tbl, self.tre, bypass_tips=True)
        actual = actual_art.view(skbio.DistanceMatrix)
        self.assertEqual(actual.ids, self.expected.ids)

    def test_accepted_types_have_consistent_behavior(self):
        tbl_rf = self.artifact('FeatureTable[RelativeFrequency]',
                               'crawford_rf.biom')

        for table in [self.tbl, tbl_rf]:
            actual_art, = self.fn(table=table, phylogeny=self.tre)
            actual = actual_art.view(skbio.DistanceMatrix)
            self.assertEqual(actual.ids, self.expected.ids)
            for id1 in actual.ids:
                for id2 in actual.ids:
                    npt.assert_almost_equal(actual[id1, id2],
                                            self.expected[id1, id2])

    def test_missing_tips_tree(self):
        tre = self.artifact('Phylogeny[Rooted]', 'root_only.tree')
        with self.assertRaises(CalledProcessError):
            obs = self.fn(self.tbl, tre)
            self.assertTrue('not a subset of the tree tips', obs.stderr)

    def test_extra_tip_tree(self):
        tbl = self.artifact('FeatureTable[Frequency]',
                            'two_feature_table.biom')

        tre_2 = self.artifact('Phylogeny[Rooted]', 'two_feature.tree')
        obs_2_art, = self.fn(tbl, tre_2)
        obs_2 = obs_2_art.view(skbio.DistanceMatrix)

        tre_3 = self.artifact('Phylogeny[Rooted]', 'three_feature.tree')
        obs_3_art, = self.fn(tbl, tre_3)
        obs_3 = obs_3_art.view(skbio.DistanceMatrix)

        self.assertEqual(obs_2, obs_3)


class BetaPhylogeneticMetaPassthroughTests(TestPluginBase):
    package = 'q2_diversity_lib.tests'

    def setUp(self):
        super().setUp()
        self.fn = self.plugin.actions['beta_phylogenetic_meta_passthrough']

        self.empty_tbl = Artifact.import_data(
            'FeatureTable[Frequency]', biom.Table(np.array([]), [], []))

        # checking parity with the unifrac.meta tests
        def unifrac_data(fn):
            path = os.path.join('data', fn)
            return pkg_resources.resource_filename('unifrac.tests', path)

        self.tables = [
            Artifact.import_data('FeatureTable[Frequency]',
                                 unifrac_data('e1.biom')),
            Artifact.import_data('FeatureTable[Frequency]',
                                 unifrac_data('e2.biom')),
        ]

        self.trees = [
            Artifact.import_data('Phylogeny[Rooted]',
                                 unifrac_data('t1.newick')),
            Artifact.import_data('Phylogeny[Rooted]',
                                 unifrac_data('t2.newick')),
        ]

    def test_method(self):
        for metric in METRICS['PHYLO']['UNIMPL']:
            obs_art, = self.fn(tables=self.tables, phylogenies=self.trees,
                               metric=metric)
            obs = obs_art.view(skbio.DistanceMatrix)
            self.assertEqual(('A', 'B', 'C'), obs.ids)
            self.assertEqual((3, 3), obs.shape)

    def test_passed_bad_metric(self):
        with self.assertRaisesRegex(TypeError,
                                    'imaginary_metric.*incompatible'):
            self.fn(tables=self.tables, phylogenies=self.trees,
                    metric='imaginary_metric')


class BetaPassthroughTests(TestPluginBase):
    package = 'q2_diversity_lib.tests'

    def setUp(self):
        super().setUp()
        self.method = self.plugin.actions['beta_passthrough']
        empty_table = biom.Table(np.array([]), [], [])
        self.empty_table = Artifact.import_data('FeatureTable[Frequency]',
                                                empty_table)
        crawford_tbl = self.get_data_path('crawford.biom')
        self.crawford_tbl = Artifact.import_data('FeatureTable[Frequency]',
                                                 crawford_tbl)
        table = biom.Table(np.array([[0, 1, 3], [1, 1, 2]]),
                           ['O1', 'O2'],
                           ['S1', 'S2', 'S3'])
        self.table = Artifact.import_data('FeatureTable[Frequency]', table)

    def test_method(self):
        for metric in METRICS['NONPHYLO']['UNIMPL']:
            self.method(table=self.crawford_tbl, metric=metric)
        # If we get here, then our methods ran without error
        self.assertTrue(True)

    def test_passed_empty_table(self):
        for metric in METRICS['NONPHYLO']['UNIMPL']:
            with self.assertRaisesRegex(ValueError, 'empty'):
                self.method(table=self.empty_table, metric=metric)

    def test_passed_bad_metric(self):
        with self.assertRaisesRegex(TypeError,
                                    'imaginary_metric.*incompatible'):
            self.method(table=self.crawford_tbl, metric='imaginary_metric')

    def test_passed_implemented_metric(self):
        # beta_passthrough does not provide access to measures that have been
        # implemented locally
        for metric in METRICS['NONPHYLO']['IMPL']:
            with self.assertRaisesRegex(TypeError, f"{metric}.*incompatible"):
                self.method(table=self.crawford_tbl, metric=metric)

    def test_aitchison(self):
        actual, = self.method(table=self.table, metric='aitchison')
        actual = actual.view(skbio.DistanceMatrix)
        expected = skbio.DistanceMatrix([[0.0000000, 0.4901290, 0.6935510],
                                         [0.4901290, 0.0000000, 0.2034219],
                                         [0.6935510, 0.2034219, 0.0000000]],
                                        ids=['S1', 'S2', 'S3'])

        self.assertEqual(actual.ids, expected.ids)
        for id1 in actual.ids:
            for id2 in actual.ids:
                npt.assert_almost_equal(actual[id1, id2], expected[id1, id2])

    def test_canberra_adkins(self):
        t = biom.Table(np.array([[0, 0], [0, 1], [1, 2]]),
                       ['O1', 'O2', 'O3'],
                       ['S1', 'S2'])
        t = Artifact.import_data('FeatureTable[Frequency]', t)
        # expected calculated by hand
        expected = skbio.DistanceMatrix(np.array([[0.0, 0.66666666666],
                                                  [0.66666666666, 0.0]]),
                                        ids=['S1', 'S2'])
        actual, = self.method(table=t, metric='canberra_adkins')
        actual = actual.view(skbio.DistanceMatrix)

        self.assertEqual(actual.ids, expected.ids)
        for id1 in actual.ids:
            for id2 in actual.ids:
                npt.assert_almost_equal(actual[id1, id2], expected[id1, id2])

    def test_beta_canberra_adkins_negative_values(self):
        t = biom.Table(np.array([[0, 0], [0, 1], [-1, -2]]),
                       ['O1', 'O2', 'O3'],
                       ['S1', 'S2'])
        t = Artifact.import_data('FeatureTable[Frequency]', t)

        with self.assertRaisesRegex(ValueError, '.*negative values'):
            self.method(table=t, metric='canberra_adkins')

    def test_jensenshannon(self):
        # expected computed with scipy.spatial.distance.jensenshannon
        expected = skbio.DistanceMatrix([[0.0000000, 0.4645014, 0.52379239],
                                         [0.4645014, 0.0000000, 0.07112939],
                                         [0.52379239, 0.07112939, 0.0000000]],
                                        ids=['S1', 'S2', 'S3'])

        actual, = self.method(table=self.table, metric='jensenshannon')
        actual = actual.view(skbio.DistanceMatrix)
        self.assertEqual(actual.ids, expected.ids)
        for id1 in actual.ids:
            for id2 in actual.ids:
                npt.assert_almost_equal(actual[id1, id2], expected[id1, id2])


class BetaPhylogeneticPassthroughTests(TestPluginBase):
    package = 'q2_diversity_lib.tests'

    def artifact(self, semantic_type, fn):
        return Artifact.import_data(semantic_type, self.get_data_path(fn))

    def setUp(self):
        super().setUp()

        self.fn = self.plugin.actions['beta_phylogenetic_passthrough']

        self.tbl = self.artifact('FeatureTable[Frequency]', 'crawford.biom')
        self.tre = self.artifact('Phylogeny[Rooted]', 'crawford.nwk')

    def test_method(self):
        for metric in METRICS['PHYLO']['UNIMPL']:
            obs_art, = self.fn(table=self.tbl, phylogeny=self.tre,
                               metric=metric)
            obs = obs_art.view(skbio.DistanceMatrix)
            self.assertEqual(9, len(obs.ids))
            self.assertEqual((9, 9), obs.shape)

        # variance adjusted
        for metric in METRICS['PHYLO']['UNIMPL']:
            obs_art, = self.fn(table=self.tbl, phylogeny=self.tre,
                               metric=metric, variance_adjusted=True)
            obs = obs_art.view(skbio.DistanceMatrix)
            self.assertEqual(9, len(obs.ids))
            self.assertEqual((9, 9), obs.shape)

        # bypass tips
        for metric in METRICS['PHYLO']['UNIMPL']:
            obs_art, = self.fn(table=self.tbl, phylogeny=self.tre,
                               metric=metric, bypass_tips=True)
            obs = obs_art.view(skbio.DistanceMatrix)
            self.assertEqual(9, len(obs.ids))
            self.assertEqual((9, 9), obs.shape)

    def test_passed_empty_table(self):
        tbl = Artifact.import_data(
            'FeatureTable[Frequency]', biom.Table(np.array([]), [], []))

        for metric in METRICS['PHYLO']['UNIMPL']:
            with self.assertRaisesRegex(ValueError, 'empty'):
                self.fn(table=tbl, phylogeny=self.tre, metric=metric)

    def test_passed_bad_metric(self):
        with self.assertRaisesRegex(TypeError, 'imaginary.*incompatible'):
            self.fn(table=self.tbl, phylogeny=self.tre, metric='imaginary')

    def test_beta_phylogenetic_alpha_on_non_generalized(self):
        with self.assertRaisesRegex(ValueError, 'The alpha parameter is only '
                                    'allowed when the selected metric is '
                                    '\'generalized_unifrac\''):
            self.fn(table=self.tbl, phylogeny=self.tre,
                    metric='unweighted_unifrac', alpha=0.11)



import numpy as np
import numpy.testing as npt

from qiime2.plugin.testing import TestPluginBase

from q2_diversity_lib.skbio._methods import (_berger_parker, _brillouin_d,
                                             _simpsons_dominance, _esty_ci,
                                             _goods_coverage, _margalef,
                                             _mcintosh_d, _strong)


class SkbioTests(TestPluginBase):
    package = 'q2_diversity_lib.skbio'

# tests for passthrough metrics were sourced from skbio
    def test_berger_parker_d(self):
        self.assertEqual(_berger_parker(np.array([5, 5])), 0.5)
        self.assertEqual(_berger_parker(np.array([1, 1, 1, 1, 0])), 0.25)

    def test_brillouin_d(self):
        self.assertAlmostEqual(_brillouin_d(np.array([1, 2, 0, 0, 3, 1])),
                               0.86289353018248782)

    def test_esty_ci(self):
        def _diversity(indices, f):
            """Calculate diversity index for each window of size 1.
            indices: vector of indices of taxa
            f: f(counts) -> diversity measure
            """
            result = []
            max_size = max(indices) + 1
            freqs = np.zeros(max_size, dtype=int)
            for i in range(len(indices)):
                freqs += np.bincount(indices[i:i + 1], minlength=max_size)
                try:
                    curr = f(freqs)
                except (ZeroDivisionError, FloatingPointError):
                    curr = 0
                result.append(curr)
            return np.array(result)

        data = [1, 1, 2, 1, 1, 3, 2, 1, 3, 4]

        observed_lower, observed_upper = zip(*_diversity(data, _esty_ci))

        expected_lower = np.array([1, -1.38590382, -0.73353593, -0.17434465,
                                   -0.15060902, -0.04386191, -0.33042054,
                                   -0.29041008, -0.43554755, -0.33385652])
        expected_upper = np.array([1, 1.38590382, 1.40020259, 0.67434465,
                                   0.55060902, 0.71052858, 0.61613483,
                                   0.54041008, 0.43554755, 0.53385652])

        npt.assert_array_almost_equal(observed_lower, expected_lower)
        npt.assert_array_almost_equal(observed_upper, expected_upper)

    def test_simpson(self):
        self.assertAlmostEqual(_simpsons_dominance(np.array([1, 0, 2, 5, 2])),
                               0.66)
        self.assertAlmostEqual(_simpsons_dominance(np.array([5])), 0)

    def test_goods_coverage(self):
        counts = [1] * 75 + [2, 2, 2, 2, 2, 2, 3, 4, 4]
        obs = _goods_coverage(counts)
        self.assertAlmostEqual(obs, 0.23469387755)

    def test_margalef(self):

        self.assertEqual(_margalef(np.array([0, 1, 1, 4, 2, 5, 2, 4, 1, 2])),
                         8 / np.log(22))

    def test_mcintosh_d(self):
        self.assertAlmostEqual(_mcintosh_d(np.array([1, 2, 3])),
                               0.636061424871458)

    def test_strong(self):
        self.assertAlmostEqual(_strong(np.array([1, 2, 3, 1])), 0.214285714)



# ----------------------------------------------------------------------------
# Copyright (c) 2018-2023, QIIME 2 development team.
#
# Distributed under the terms of the Modified BSD License.
#
# The full license is in the file LICENSE, distributed with this software.
# ----------------------------------------------------------------------------

from subprocess import CalledProcessError

import numpy as np
import numpy.testing as npt
import pandas as pd
import pandas.testing as pdt
import biom

from qiime2.plugin.testing import TestPluginBase
from qiime2 import Artifact

from ..alpha import (pielou_evenness, observed_features,
                     shannon_entropy, METRICS,
                     _berger_parker, _brillouin_d,
                     _simpsons_dominance, _esty_ci,
                     _goods_coverage, _margalef,
                     _mcintosh_d, _strong
                     )


class SmokeTests(TestPluginBase):
    package = 'q2_diversity_lib.tests'

    def setUp(self):
        super().setUp()
        empty_table = biom.Table(np.array([]), [], [])
        self.empty_table = Artifact.import_data('FeatureTable[Frequency]',
                                                empty_table)

    def test_non_phylogenetic_passed_empty_table(self):
        for metric in METRICS['NONPHYLO']['IMPL']:
            metric = METRICS['NAME_TRANSLATIONS'][metric]
            with self.assertRaisesRegex(ValueError, 'empty'):
                self.plugin.actions[metric](table=self.empty_table)


class FaithPDTests(TestPluginBase):
    package = 'q2_diversity_lib.tests'

    def artifact(self, semantic_type, fp):
        return Artifact.import_data(semantic_type, self.get_data_path(fp))

    @staticmethod
    def assert_vec_equal(actual_art, expected):
        actual = actual_art.view(pd.Series)
        pdt.assert_series_equal(actual, expected)

    def setUp(self):
        super().setUp()
        self.fn = self.plugin.actions['faith_pd']
        self.tbl = self.artifact('FeatureTable[Frequency]',
                                 'faith_test_table.biom')
        self.tre = self.artifact('Phylogeny[Rooted]', 'faith_test.tree')
        self.expected = pd.Series({'S1': 0.5, 'S2': 0.7, 'S3': 1.0,
                                   'S4': 100.5, 'S5': 101},
                                  name='faith_pd')

    def test_receives_empty_table(self):
        table = self.artifact('FeatureTable[Frequency]', 'empty_table.biom')
        with self.assertRaisesRegex(ValueError, 'empty'):
            self.fn(table=table, phylogeny=self.tre)

    def test_method(self):
        actual_art, = self.fn(table=self.tbl, phylogeny=self.tre)
        self.assert_vec_equal(actual_art, self.expected)

    def test_accepted_types_have_consistent_behavior(self):
        rf_tbl = self.artifact('FeatureTable[RelativeFrequency]',
                               'faith_test_table_rf.biom')
        pa_tbl = self.artifact('FeatureTable[PresenceAbsence]',
                               'faith_test_table_pa.biom')
        for table in [self.tbl, rf_tbl, pa_tbl]:
            actual_art, = self.fn(table=table, phylogeny=self.tre)
            self.assert_vec_equal(actual_art, self.expected)

    def test_passed_tree_missing_tip(self):
        tree = self.artifact('Phylogeny[Rooted]', 'missing_tip.tree')
        with self.assertRaises(CalledProcessError):
            obs = self.fn(table=self.tbl, phylogeny=tree)
            self.assertTrue('not a subset of the tree tips' in obs.stderr)

    def test_passed_rootonlytree(self):
        tree = self.artifact('Phylogeny[Rooted]', 'root_only.tree')
        with self.assertRaises(CalledProcessError):
            obs = self.fn(table=self.tbl, phylogeny=tree)
            self.assertTrue('not a subset of the tree tips' in obs.stderr)


class ObservedFeaturesTests(TestPluginBase):
    package = 'q2_diversity_lib.tests'

    def setUp(self):
        super().setUp()
        self.input_table = biom.Table(np.array([[1, 0, .5, 999, 1],
                                                [0, 1, 2, 0, 5],
                                                [0, 0, 0, 1, 10]]),
                                      ['A', 'B', 'C'],
                                      ['S1', 'S2', 'S3', 'S4', 'S5'])
        # Calculated by hand:
        self.expected = pd.Series(
                {'S1': 1, 'S2': 1, 'S3': 2, 'S4': 2,
                 'S5': 3},
                name='observed_features')

    def test_method(self):
        actual = observed_features(table=self.input_table)
        pdt.assert_series_equal(actual, self.expected)

    def test_accepted_types_have_consistent_behavior(self):
        freq_table = self.input_table
        rel_freq_table = self.input_table.norm(axis='sample',
                                               inplace=False)
        p_a_table = self.input_table.pa(inplace=False)
        accepted_tables = [freq_table, rel_freq_table, p_a_table]
        for table in accepted_tables:
            actual = observed_features(table)
            pdt.assert_series_equal(actual, self.expected)


class PielouEvennessTests(TestPluginBase):
    package = 'q2_diversity_lib.tests'

    def setUp(self):
        super().setUp()
        self.input_table = biom.Table(np.array([[0, 1, 1, 1, 999, 1],
                                                [0, 0, 1, 1, 999, 1],
                                                [0, 0, 0, 1, 999, 2]]),
                                      ['A', 'B', 'C'],
                                      ['S1', 'S2', 'S3', 'S4', 'S5', 'S6'])
        # Calculated by hand:
        self.expected = pd.Series(
                {'S1': np.NaN, 'S2': np.NaN, 'S3': 1, 'S4': 1,
                 'S5': 1, 'S6': 0.946394630357186},
                name='pielou_evenness')

    def test_method(self):
        actual = pielou_evenness(table=self.input_table)
        pdt.assert_series_equal(actual, self.expected)

    def test_accepted_types_have_consistent_behavior(self):
        freq_table = self.input_table
        rel_freq_table = self.input_table.norm(axis='sample',
                                               inplace=False)
        accepted_tables = [freq_table, rel_freq_table]
        for table in accepted_tables:
            actual = pielou_evenness(table)
            pdt.assert_series_equal(actual, self.expected)

    def test_drop_undefined_samples(self):
        NaN_table = biom.Table(np.array([[0, 1, 0, 0, 1, 1],
                                         [0, 0, 1, 0, 1, 1],
                                         [0, 0, 0, 1, 0, 1]]),
                               ['A', 'B', 'C'],
                               ['S1', 'S2', 'S3', 'S4', 'S5', 'S6'])
        # pandas supports floating point correction for float dtype only,
        # these 1 ints were changed to 1.0 floats to get that support
        expected = pd.Series({'S5': 1.0, 'S6': 1.0}, name='pielou_evenness')
        actual = pielou_evenness(table=NaN_table, drop_undefined_samples=True)
        pdt.assert_series_equal(actual, expected, check_dtype=False)

    def test_do_not_drop_undefined_samples(self):
        NaN_table = biom.Table(np.array([[0, 1, 0, 0, 1, 1],
                                         [0, 0, 1, 0, 1, 1],
                                         [0, 0, 0, 1, 0, 1]]),
                               ['A', 'B', 'C'],
                               ['S1', 'S2', 'S3', 'S4', 'S5', 'S6'])
        expected = pd.Series({'S1': np.NaN, 'S2': np.NaN, 'S3': np.NaN,
                              'S4': np.NaN, 'S5': 1, 'S6': 1},
                             name='pielou_evenness')
        actual = pielou_evenness(table=NaN_table, drop_undefined_samples=False)
        pdt.assert_series_equal(actual, expected)


class ShannonEntropyTests(TestPluginBase):
    package = 'q2_diversity_lib.tests'

    def setUp(self):
        super().setUp()
        self.input_table = biom.Table(np.array([[0, 0, 0, 0, 1, 1],
                                                [0, 1, 0, 0, 1, 1],
                                                [0, 0, 1, 0, 1, 1],
                                                [0, 0, 1, 0, 0, 1]]),
                                      ['A', 'B', 'C', 'D'],
                                      ['S1', 'S2', 'S3', 'S4', 'S5', 'S6'])
        # Calculated by hand:
        self.expected = pd.Series(
                {'S1': np.NaN, 'S2': 0, 'S3': 1, 'S4': np.NaN,
                 'S5': 1.584962500721156, 'S6': 2},
                name='shannon_entropy')

    def test_method(self):
        actual = shannon_entropy(table=self.input_table)
        pdt.assert_series_equal(actual, self.expected)

    def test_accepted_types_have_consistent_behavior(self):
        freq_table = self.input_table
        rel_freq_table = self.input_table.norm(axis='sample',
                                               inplace=False)
        accepted_tables = [freq_table, rel_freq_table]
        for table in accepted_tables:
            actual = shannon_entropy(table)
            pdt.assert_series_equal(actual, self.expected)

    def test_drop_undefined_samples(self):
        expected = pd.Series({'S2': 0, 'S3': 1, 'S5': 1.584962500721156,
                              'S6': 2}, name='shannon_entropy')
        actual = shannon_entropy(table=self.input_table,
                                 drop_undefined_samples=True)
        pdt.assert_series_equal(actual, expected, check_dtype=False)

    def test_do_not_drop_undefined_samples(self):
        actual = shannon_entropy(table=self.input_table,
                                 drop_undefined_samples=False)
        pdt.assert_series_equal(actual, self.expected)


class AlphaPassthroughTests(TestPluginBase):
    package = 'q2_diversity_lib.tests'

    def setUp(self):
        super().setUp()
        self.method = self.plugin.actions['alpha_passthrough']
        empty_table = biom.Table(np.array([]), [], [])
        self.empty_table = Artifact.import_data('FeatureTable[Frequency]',
                                                empty_table)
        crawford_tbl = self.get_data_path('crawford.biom')
        self.crawford_tbl = Artifact.import_data('FeatureTable[Frequency]',
                                                 crawford_tbl)

    def test_method(self):
        for metric in METRICS['NONPHYLO']['UNIMPL']:
            # NOTE: crawford table used b/c input_table too minimal for `ace`
            self.method(table=self.crawford_tbl, metric=metric)
        # If we get here, then our methods ran without error
        self.assertTrue(True)

    def test_passed_empty_table(self):
        for metric in METRICS['NONPHYLO']['UNIMPL']:
            with self.assertRaisesRegex(ValueError, 'empty'):
                self.method(table=self.empty_table, metric=metric)

    def test_passed_bad_metric(self):
        with self.assertRaisesRegex(TypeError,
                                    'imaginary_metric.*incompatible'):
            self.method(table=self.crawford_tbl, metric='imaginary_metric')

    def test_passed_implemented_metric(self):
        # alpha_passthrough does not provide access to measures that have been
        # implemented locally
        for metric in METRICS['NONPHYLO']['IMPL']:
            with self.assertRaisesRegex(TypeError, f"{metric}.*incompatible"):
                self.method(table=self.crawford_tbl, metric=metric)

# tests for passthrough metrics were sourced from skbio
    def test_berger_parker_d(self):
        self.assertEqual(_berger_parker(np.array([5, 5])), 0.5)
        self.assertEqual(_berger_parker(np.array([1, 1, 1, 1, 0])), 0.25)

    def test_brillouin_d(self):
        self.assertAlmostEqual(_brillouin_d(np.array([1, 2, 0, 0, 3, 1])),
                               0.86289353018248782)

    def test_esty_ci(self):
        def _diversity(indices, f):
            """Calculate diversity index for each window of size 1.

            indices: vector of indices of taxa
            f: f(counts) -> diversity measure

            """
            result = []
            max_size = max(indices) + 1
            freqs = np.zeros(max_size, dtype=int)
            for i in range(len(indices)):
                freqs += np.bincount(indices[i:i + 1], minlength=max_size)
                try:
                    curr = f(freqs)
                except (ZeroDivisionError, FloatingPointError):
                    curr = 0
                result.append(curr)
            return np.array(result)

        data = [1, 1, 2, 1, 1, 3, 2, 1, 3, 4]

        observed_lower, observed_upper = zip(*_diversity(data, _esty_ci))

        expected_lower = np.array([1, -1.38590382, -0.73353593, -0.17434465,
                                   -0.15060902, -0.04386191, -0.33042054,
                                   -0.29041008, -0.43554755, -0.33385652])
        expected_upper = np.array([1, 1.38590382, 1.40020259, 0.67434465,
                                   0.55060902, 0.71052858, 0.61613483,
                                   0.54041008, 0.43554755, 0.53385652])

        npt.assert_array_almost_equal(observed_lower, expected_lower)
        npt.assert_array_almost_equal(observed_upper, expected_upper)

    def test_simpson(self):
        self.assertAlmostEqual(_simpsons_dominance(np.array([1, 0, 2, 5, 2])),
                               0.66)
        self.assertAlmostEqual(_simpsons_dominance(np.array([5])), 0)

    def test_goods_coverage(self):
        counts = [1] * 75 + [2, 2, 2, 2, 2, 2, 3, 4, 4]
        obs = _goods_coverage(counts)
        self.assertAlmostEqual(obs, 0.23469387755)

    def test_margalef(self):

        self.assertEqual(_margalef(np.array([0, 1, 1, 4, 2, 5, 2, 4, 1, 2])),
                         8 / np.log(22))

    def test_mcintosh_d(self):
        self.assertAlmostEqual(_mcintosh_d(np.array([1, 2, 3])),
                               0.636061424871458)

    def test_strong(self):
        self.assertAlmostEqual(_strong(np.array([1, 2, 3, 1])), 0.214285714)



# ----------------------------------------------------------------------------
# Copyright (c) 2018-2023, QIIME 2 development team.
#
# Distributed under the terms of the Modified BSD License.
#
# The full license is in the file LICENSE, distributed with this software.
# ----------------------------------------------------------------------------

from inspect import signature
from os import environ
import subprocess

import numpy as np
from decorator import decorator
import psutil
import biom

from q2_types.feature_table import BIOMV210Format


def _drop_undefined_samples(table, minimum_nonzero_elements):
    def f(v, i, m):
        return (v > 0).sum() >= minimum_nonzero_elements
    return table.filter(f, inplace=False)


def _partition(table, block_size=100):
    number_of_splits = max(1, np.ceil(len(table.ids()) / block_size))
    splits = np.array_split(table.ids(), number_of_splits)
    split_map = {}
    for idx, split in enumerate(splits):
        for id_ in split:
            split_map[id_] = idx

    def part_f(i, m):
        return split_map[i]

    for _, block in table.partition(part_f):
        yield block


@decorator
def _validate_tables(wrapped_function, *args, **kwargs):
    bound_arguments = signature(wrapped_function).bind(*args, **kwargs)
    table = bound_arguments.arguments.get('table')
    if table is None:
        table = bound_arguments.arguments.get('tables')

    if table is None:
        raise TypeError("The wrapped function is missing argument 'table' or "
                        "'tables'")

    if not isinstance(table, (tuple, list, set)):
        table = [table]

    for tab in table:
        if isinstance(tab, BIOMV210Format):
            tab = str(tab)
            tab_obj = biom.load_table(tab)
        elif isinstance(tab, biom.Table):
            tab_obj = tab
        else:
            raise ValueError("Invalid view type: table passed as "
                             f"{type(tab)}")

        if tab_obj.is_empty():
            raise ValueError("The provided table is empty")

        if np.isnan(tab_obj.matrix_data.data).sum() > 0:
            raise ValueError("The provided table contains NaN")

        if (tab_obj.matrix_data.data < 0).sum() > 0:
            raise ValueError("The provided table contains negative values")

    return wrapped_function(*args, **kwargs)


@decorator
def _validate_requested_cpus(wrapped_function, *args, **kwargs):
    bound_arguments = signature(wrapped_function).bind(*args, **kwargs)
    bound_arguments.apply_defaults()
    b_a_arguments = bound_arguments.arguments

    if 'n_jobs' in b_a_arguments and 'threads' in b_a_arguments:
        raise TypeError("Duplicate parameters: The _validate_requested_cpus "
                        "decorator may not be applied to callables with both "
                        "'n_jobs' and 'threads' parameters. Do you really need"
                        " both?")
    elif 'n_jobs' in b_a_arguments:
        param_name = 'n_jobs'
    elif 'threads' in b_a_arguments:
        param_name = 'threads'
    else:
        raise TypeError("The _validate_requested_cpus decorator may not be"
                        " applied to callables without an 'n_jobs' or "
                        "'threads' parameter.")

    # If `Process.cpu_affinity` unavailable on system, fall back
    # https://psutil.readthedocs.io/en/latest/index.html#psutil.cpu_count
    try:
        cpus_available = len(psutil.Process().cpu_affinity())
    except AttributeError:
        cpus_available = psutil.cpu_count(logical=False)

    cpus_requested = b_a_arguments[param_name]

    if cpus_requested == 0:
        # mutate bound_arguments.arguments 'auto' to the requested # of cpus...
        b_a_arguments[param_name] = cpus_available
        # ...and update cpus requested to prevent TypeError
        cpus_requested = cpus_available

    if cpus_requested > cpus_available:
        raise ValueError(f"The value passed to '{param_name}' cannot exceed "
                         f"the number of processors ({cpus_available}) "
                         "available to the system.")

    return wrapped_function(*bound_arguments.args, **bound_arguments.kwargs)


def _run_external_cmd(cmd, verbose=True, env=None):
    if verbose:
        print("Running external command line application. This may print"
              " messages to stdout and/or stderr.\nThe command being run is"
              " below. This command cannot be manually re-run as it will"
              " depend on temporary files that no longer exist.\n\nCommand:\n")
        print(" ".join(cmd), end='\n\n')
    return subprocess.run(cmd, check=True, env=env)


def _omp_cmd_wrapper(threads, cmd, verbose=True):
    env = environ.copy()
    env.update({'OMP_NUM_THREADS': str(threads)})
    return _run_external_cmd(cmd, verbose=verbose, env=env)



# ----------------------------------------------------------------------------
# Copyright (c) 2018-2023, QIIME 2 development team.
#
# Distributed under the terms of the Modified BSD License.
#
# The full license is in the file LICENSE, distributed with this software.
# ----------------------------------------------------------------------------

import biom
import skbio.diversity
import sklearn.metrics
import unifrac
from skbio.stats.composition import clr
from scipy.spatial.distance import euclidean
from scipy.spatial.distance import jensenshannon
import numpy as np

from q2_types.distance_matrix import LSMatFormat
from q2_types.feature_table import BIOMV210Format
from q2_types.tree import NewickFormat

from ._util import (_validate_tables,
                    _validate_requested_cpus,
                    _omp_cmd_wrapper)


# NOTE: some phylo metrics are currently in both implemented and unimplemented
# collections, when implemented with certain params only (e.g. both 'vanilla'
# and Variance Adjusted weighted unifracs use unifrac.weighted_unnormalized,
# but only 'vanilla' is currently implemented)
METRICS = {
    'PHYLO': {
        'IMPL': {'unweighted_unifrac', 'weighted_unifrac'},
        'UNIMPL': {'unweighted_unifrac', 'weighted_unifrac',
                   'weighted_normalized_unifrac', 'generalized_unifrac'},
    },
    'NONPHYLO': {
        'IMPL': {'braycurtis', 'jaccard'},
        'UNIMPL': {'cityblock', 'euclidean', 'seuclidean', 'sqeuclidean',
                   'cosine', 'correlation', 'hamming', 'chebyshev', 'canberra',
                   'yule', 'matching', 'dice',
                   'rogerstanimoto', 'russellrao', 'sokalmichener',
                   'sokalsneath', 'minkowski', 'aitchison', 'canberra_adkins',
                   'jensenshannon'}
    },
    'NAME_TRANSLATIONS': {'braycurtis': 'bray_curtis',
                          'jaccard': 'jaccard',
                          'unweighted_unifrac': 'unweighted_unifrac',
                          'weighted_unifrac': 'weighted_unifrac',
                          }
}


# -------------------- Method Dispatch -----------------------
@_validate_tables
@_validate_requested_cpus
def beta_passthrough(table: biom.Table, metric: str, pseudocount: int = 1,
                     n_jobs: int = 1) -> skbio.DistanceMatrix:
    def aitchison(x, y, **kwds):
        return euclidean(clr(x), clr(y))

    def canberra_adkins(x, y, **kwds):
        nz = ((x > 0) | (y > 0))
        x_ = x[nz]
        y_ = y[nz]
        nnz = nz.sum()

        return (1. / nnz) * np.sum(np.abs(x_ - y_) / (x_ + y_))

    def jensen_shannon(x, y, **kwds):
        return jensenshannon(x, y)

    counts = table.matrix_data.toarray().T
    sample_ids = table.ids(axis='sample')
    if metric == 'aitchison':
        counts += pseudocount
        metric = aitchison
    elif metric == 'canberra_adkins':
        metric = canberra_adkins
    elif metric == 'jensenshannon':
        metric = jensen_shannon
    else:
        pass

    return skbio.diversity.beta_diversity(
            metric=metric, counts=counts, ids=sample_ids, validate=False,
            pairwise_func=sklearn.metrics.pairwise_distances, n_jobs=n_jobs)


@_validate_tables
@_validate_requested_cpus
def beta_phylogenetic_passthrough(table: BIOMV210Format,
                                  phylogeny: NewickFormat,
                                  metric: str,
                                  threads: int = 1,
                                  variance_adjusted: bool = False,
                                  alpha: float = None,
                                  bypass_tips: bool = False
                                  ) -> LSMatFormat:
    # Ideally we remove this when we can support optional type-mapped params.
    if alpha is not None and metric != 'generalized_unifrac':
        raise ValueError("The alpha parameter is only allowed when the "
                         "selected metric is 'generalized_unifrac'")

    method = {
        'unweighted_unifrac': 'unweighted',
        'weighted_unifrac': 'weighted_unnormalized',
        'weighted_normalized_unifrac': 'weighted_normalized',
        'generalized_unifrac': 'generalized',
    }[metric]

    result = LSMatFormat()

    cmd = [
        'ssu',
        '-i', str(table),
        '-t', str(phylogeny),
        '-m', method,
        '-o', str(result),
    ]

    # handle unimplemented unifracs
    if metric == 'generalized_unifrac':
        alpha = 1.0 if alpha is None else alpha
        cmd += ['-a', str(alpha)]

    if variance_adjusted:
        cmd += ['--vaw']

    if bypass_tips:
        cmd += ['-f']

    _omp_cmd_wrapper(threads, cmd)

    return result


@_validate_tables
@_validate_requested_cpus
def beta_phylogenetic_meta_passthrough(tables: BIOMV210Format,
                                       phylogenies: NewickFormat,
                                       metric: str,
                                       threads: int = 1,
                                       variance_adjusted: bool = False,
                                       alpha: float = None,
                                       bypass_tips: bool = False,
                                       weights: list = None,
                                       consolidation: str='skipping_missing_values'  # noqa
                                       ) -> skbio.DistanceMatrix:
    # Ideally we remove this when we can support optional type-mapped params.
    if alpha is not None and metric != 'generalized_unifrac':
        raise ValueError("The alpha parameter is only allowed when the "
                         "selected metric is 'generalized_unifrac'")

    metric_map = {'unweighted_unifrac': 'unweighted',
                  'weighted_normalized_unifrac': 'weighted_normalized',
                  'weighted_unifrac': 'weighted_unnormalized',
                  'generalized_unifrac': 'generalized'}
    metric = metric_map[metric]

    return unifrac.meta(tuple([str(t) for t in tables]),
                        tuple([str(p) for p in phylogenies]),
                        weights=weights, threads=threads,
                        consolidation=consolidation, method=metric,
                        variance_adjusted=variance_adjusted,
                        alpha=alpha, bypass_tips=bypass_tips)


# --------------------Non-Phylogenetic-----------------------
@_validate_tables
@_validate_requested_cpus
def bray_curtis(table: biom.Table, n_jobs: int = 1) -> skbio.DistanceMatrix:
    counts = table.matrix_data.toarray().T
    sample_ids = table.ids(axis='sample')
    return skbio.diversity.beta_diversity(
        metric='braycurtis',
        counts=counts,
        ids=sample_ids,
        validate=False,
        pairwise_func=sklearn.metrics.pairwise_distances,
        n_jobs=n_jobs
    )


@_validate_tables
@_validate_requested_cpus
def jaccard(table: biom.Table, n_jobs: int = 1) -> skbio.DistanceMatrix:
    counts = table.matrix_data.toarray().T
    sample_ids = table.ids(axis='sample')
    return skbio.diversity.beta_diversity(
        metric='jaccard',
        counts=counts,
        ids=sample_ids,
        validate=False,
        pairwise_func=sklearn.metrics.pairwise_distances,
        n_jobs=n_jobs
    )


# ------------------------Phylogenetic-----------------------
@_validate_tables
@_validate_requested_cpus
def unweighted_unifrac(table: BIOMV210Format,
                       phylogeny: NewickFormat,
                       threads: int = 1,
                       bypass_tips: bool = False) -> LSMatFormat:
    result = LSMatFormat()

    cmd = [
        'ssu',
        '-i', str(table),
        '-t', str(phylogeny),
        '-m', 'unweighted',
        '-o', str(result),
    ]

    if bypass_tips:
        cmd += ['-f']

    _omp_cmd_wrapper(threads, cmd)

    return result


@_validate_tables
@_validate_requested_cpus
def weighted_unifrac(table: BIOMV210Format, phylogeny: NewickFormat,
                     threads: int = 1, bypass_tips: bool = False
                     ) -> LSMatFormat:
    result = LSMatFormat()

    cmd = [
        'ssu',
        '-i', str(table),
        '-t', str(phylogeny),
        '-m', 'weighted_unnormalized',
        '-o', str(result),
    ]

    if bypass_tips:
        cmd += ['-f']

    _omp_cmd_wrapper(threads, cmd)

    return result



# ----------------------------------------------------------------------------
# Copyright (c) 2018-2023, QIIME 2 development team.
#
# Distributed under the terms of the Modified BSD License.
#
# The full license is in the file LICENSE, distributed with this software.
# ----------------------------------------------------------------------------

import pandas as pd
import numpy as np
import functools

import skbio.diversity
import skbio.diversity.alpha

import biom

from q2_types.feature_table import BIOMV210Format
from q2_types.sample_data import AlphaDiversityFormat
from q2_types.tree import NewickFormat

from ._util import (_validate_tables,
                    _validate_requested_cpus,
                    _omp_cmd_wrapper)

from q2_diversity_lib.skbio._methods import (_berger_parker, _brillouin_d,
                                             _simpsons_dominance, _esty_ci,
                                             _goods_coverage, _margalef,
                                             _mcintosh_d, _strong, _shannon,
                                             _p_evenness)

METRICS = {
    'PHYLO': {
        'IMPL': {'faith_pd'},
        'UNIMPL': set()
    },
    'NONPHYLO': {
        'IMPL': {'observed_features', 'pielou_e', 'shannon'},
        'UNIMPL': {'ace', 'chao1', 'chao1_ci', 'berger_parker_d',
                   'brillouin_d', 'dominance', 'doubles', 'enspie', 'esty_ci',
                   'fisher_alpha', 'goods_coverage', 'heip_e',
                   'kempton_taylor_q', 'margalef', 'mcintosh_d', 'mcintosh_e',
                   'menhinick', 'michaelis_menten_fit', 'osd', 'robbins',
                   'simpson', 'simpson_e', 'singles', 'strong', 'gini_index',
                   'lladser_pe'
                   }
    },
    'NAME_TRANSLATIONS': {'faith_pd': 'faith_pd',
                          'shannon': 'shannon_entropy',
                          'pielou_e': 'pielou_evenness',
                          'observed_features': 'observed_features'
                          }
}


# --------------------- Phylogenetic -----------------------------------------
@_validate_tables
@_validate_requested_cpus
def faith_pd(table: BIOMV210Format, phylogeny: NewickFormat,
             threads: int = 1) -> AlphaDiversityFormat:
    vec = AlphaDiversityFormat()
    cmd = ['faithpd', '-i', str(table), '-t', str(phylogeny), '-o', str(vec)]
    _omp_cmd_wrapper(threads, cmd)
    return vec


# --------------------- Non-Phylogenetic -------------------------------------
def _skbio_alpha_diversity_from_1d(v, metric):
    # alpha_diversity expects a 2d structure
    v = np.reshape(v, (1, len(v)))
    result = skbio.diversity.alpha_diversity(metric=metric,
                                             counts=v,
                                             ids=['placeholder', ],
                                             validate=False)
    return result.iloc[0]


@_validate_tables
def observed_features(table: biom.Table) -> pd.Series:
    presence_absence_table = table.pa(inplace=False)
    results = []
    for v in presence_absence_table.iter_data(dense=True):
        results.append(_skbio_alpha_diversity_from_1d(v.astype(int),
                                                      'observed_otus'))
    results = pd.Series(results, index=table.ids(), name='observed_features')
    return results


@_validate_tables
def pielou_evenness(table: biom.Table,
                    drop_undefined_samples: bool = False) -> pd.Series:
    if drop_undefined_samples:
        def transform_(v, i, m):
            if (v > 0).sum() < 2:
                return np.zeros(len(v))
            else:
                return v

        table = table.transform(transform_, inplace=False).remove_empty()

    results = []
    for v in table.iter_data(dense=True):
        # using in-house metrics temporarily
        # results.append(_skbio_alpha_diversity_from_1d(v, 'pielou_e'))
        v = np.reshape(v, (1, len(v)))
        results.extend([_p_evenness(c)for c in v])
    results = pd.Series(results, index=table.ids(), name='pielou_evenness')
    return results


@_validate_tables
def shannon_entropy(table: biom.Table,
                    drop_undefined_samples: bool = False) -> pd.Series:
    if drop_undefined_samples:
        table = table.remove_empty(inplace=False)

    results = []
    for v in table.iter_data(dense=True):
        # using in-house metrics temporarily
        # results.append(_skbio_alpha_diversity_from_1d(v, 'shannon'))
        v = np.reshape(v, (1, len(v)))
        results.extend([_shannon(c)for c in v])
    results = pd.Series(results, index=table.ids(), name='shannon_entropy')
    return results


@_validate_tables
def alpha_passthrough(table: biom.Table, metric: str) -> pd.Series:
    results = []
    method_map = {"berger_parker_d": _berger_parker,
                  "brillouin_d": _brillouin_d,
                  "simpson": _simpsons_dominance,
                  "esty_ci": _esty_ci,
                  "goods_coverage": _goods_coverage,
                  "margalef": _margalef,
                  "mcintosh_d": _mcintosh_d,
                  "strong": _strong}

    if metric in method_map:
        metric = functools.partial(method_map[metric])
        for v in table.iter_data(dense=True):
            v = np.reshape(v, (1, len(v)))
            results.extend([metric(c)for c in v])
    else:
        for v in table.iter_data(dense=True):
            results.append(_skbio_alpha_diversity_from_1d(v.astype(int),
                                                          metric))
    results = pd.Series(results, index=table.ids(), name=metric)
    return results



# ----------------------------------------------------------------------------
# Copyright (c) 2018-2023, QIIME 2 development team.
#
# Distributed under the terms of the Modified BSD License.
#
# The full license is in the file LICENSE, distributed with this software.
# ----------------------------------------------------------------------------

from qiime2.plugin import (Plugin, Citations, Bool, Int, Range, Choices, Str,
                           Float, List, Threads)
from q2_types.feature_table import (FeatureTable, Frequency, RelativeFrequency,
                                    PresenceAbsence)
from q2_types.tree import Phylogeny, Rooted
from q2_types.sample_data import AlphaDiversity, SampleData
from q2_types.distance_matrix import DistanceMatrix
from unifrac._meta import CONSOLIDATIONS

from . import alpha, beta, __version__, examples

citations = Citations.load('citations.bib', package='q2_diversity_lib')
plugin = Plugin(
    name="diversity-lib",
    version=__version__,
    website="https://github.com/qiime2/q2-diversity-lib",
    short_description="Plugin for computing community diversity.",
    package="q2_diversity_lib",
    description="This QIIME 2 plugin computes individual metrics for "
    " community alpha and beta diversity.",
)

n_jobs_description = (
    "The number of concurrent jobs to use in performing this calculation. "
    "May not exceed the number of available physical cores. If n_jobs = "
    "'auto', one job will be launched for each identified CPU core on the "
    "host."
)

threads_description = (
    "The number of CPU threads to use in performing this calculation. "
    "May not exceed the number of available physical cores. If threads = "
    "'auto', one thread will be created for each identified CPU core on the "
    "host."
)

drop_undef_samples_description = (
    "When calculating some metrics, samples with fewer than a predetermined "
    "number of features produce undefined (NaN) values. If true, affected "
    "samples are dropped from metrics with 'drop_undefined_samples' "
    "implemented. For metrics without a 'drop_undefined_samples' parameter, "
    "this value will be ignored and no samples will be dropped."
)

# ------------------------ alpha-diversity -----------------------
plugin.methods.register_function(
    function=alpha.faith_pd,
    inputs={'table': FeatureTable[Frequency | RelativeFrequency
            | PresenceAbsence],
            'phylogeny': Phylogeny[Rooted]},
    parameters={'threads': Threads},
    outputs=[('vector', SampleData[AlphaDiversity])],
    input_descriptions={
        'table': "The feature table containing the samples for which Faith's "
                 "phylogenetic diversity should be computed. Table values "
                 "will be converted to presence/absence.",
        'phylogeny': "Phylogenetic tree containing tip identifiers that "
                     "correspond to the feature identifiers in the table. "
                     "This tree can contain tip ids that are not present in "
                     "the table, but all feature ids in the table must be "
                     "present in this tree."},
    parameter_descriptions={'threads': threads_description},
    output_descriptions={'vector': "Vector containing per-sample values for "
                                   "Faith's Phylogenetic Diversity."},
    name="Faith's Phylogenetic Diversity",
    description="Computes Faith's Phylogenetic Diversity for all samples in "
                "a feature table.",
    examples={'basic': examples.faith_pd_example},
    citations=[citations['faith1992conservation'],
               citations['armstrong2021faithspd']]
)

plugin.methods.register_function(
    function=alpha.observed_features,
    inputs={'table': FeatureTable[Frequency | RelativeFrequency
            | PresenceAbsence]},
    parameters=None,
    outputs=[('vector', SampleData[AlphaDiversity])],
    input_descriptions={'table': "The feature table containing the samples "
                        "for which the number of observed features should be "
                        "calculated. Table values will be converted to "
                        "presence/absence."},
    parameter_descriptions=None,
    output_descriptions={'vector': "Vector containing per-sample counts of "
                                   "observed features."},
    name="Observed Features",
    description="Compute the number of observed features for each sample in a "
                "feature table",
    examples={'basic': examples.observed_features_example}
)

plugin.methods.register_function(
    function=alpha.pielou_evenness,
    inputs={'table': FeatureTable[Frequency | RelativeFrequency]},
    parameters={'drop_undefined_samples': Bool},
    outputs=[('vector', SampleData[AlphaDiversity])],
    input_descriptions={'table': "The feature table containing the samples "
                        "for which Pielou's evenness should be computed."},
    parameter_descriptions={'drop_undefined_samples': "Samples with fewer than"
                            " two observed features produce undefined (NaN) "
                            "values. If true, these samples are dropped "
                            "from the output vector."},
    output_descriptions={'vector': "Vector containing per-sample values "
                                   "for Pielou's Evenness."},
    name="Pielou's Evenness",
    description="Compute Pielou's Evenness for each sample in a "
                "feature table",
    examples={'basic': examples.pielou_evenness_example,
              'dropping undefined samples': examples.pielou_drop_example},
    citations=[citations['pielou1966measurement']]
)

# TODO: Augment citations as needed
plugin.methods.register_function(
    function=alpha.shannon_entropy,
    inputs={'table': FeatureTable[Frequency | RelativeFrequency]},
    parameters={'drop_undefined_samples': Bool},
    outputs=[('vector', SampleData[AlphaDiversity])],
    input_descriptions={'table': "The feature table containing the samples "
                        "for which Shannon's Entropy should be computed."},
    parameter_descriptions={'drop_undefined_samples': "Samples with no "
                            "observed features produce undefined (NaN) values."
                            " If true, these samples are dropped from the "
                            "output vector."},
    output_descriptions={'vector': "Vector containing per-sample values "
                                   "for Shannon's Entropy."},
    name="Shannon's Entropy",
    description="Compute Shannon's Entropy for each sample in a "
                "feature table",
    examples={'basic': examples.shannon_entropy_example,
              'dropping undefined samples': examples.shannon_drop_example},
    citations=[citations['shannon1948communication']]
)


# ------------------------ beta-diversity -----------------------
# TODO: Augment citations as needed
plugin.methods.register_function(
    function=beta.bray_curtis,
    inputs={'table': FeatureTable[Frequency | RelativeFrequency]},
    parameters={'n_jobs': Threads},
    outputs=[('distance_matrix', DistanceMatrix)],
    input_descriptions={
        'table': "The feature table containing the samples for which "
                 "Bray-Curtis dissimilarity should be computed."},
    parameter_descriptions={'n_jobs': n_jobs_description},
    output_descriptions={
        'distance_matrix': "Distance matrix for Bray-Curtis dissimilarity"},
    name="Bray-Curtis Dissimilarity",
    description="Compute Bray-Curtis dissimilarity for each sample in a "
                "feature table. Note: Frequency and relative frequency data "
                "produce different results unless overall sample sizes are "
                "identical. Please consider the impact on your results if "
                "you use Bray-Curtis with count data that has not been "
                "adjusted (normalized).",
    examples={"run on one core (by default)": examples.bray_curtis_example,
              "to run on n cores, replace 1 here with your preferred integer":
              examples.bray_curtis_n_jobs_example,
              "use 'auto' to run on all of host system's available CPU cores":
              examples.bray_curtis_auto_jobs_example,
              },
    citations=[citations['sorensen1948method']])

# TODO: Augment citations as needed
plugin.methods.register_function(
    function=beta.jaccard,
    inputs={'table': FeatureTable[Frequency | RelativeFrequency
            | PresenceAbsence]},
    parameters={'n_jobs': Threads},
    outputs=[('distance_matrix', DistanceMatrix)],
    input_descriptions={
        'table': "The feature table containing the samples for which "
                 "Jaccard distance should be computed."},
    parameter_descriptions={'n_jobs': n_jobs_description},
    output_descriptions={
        'distance_matrix': "Distance matrix for Jaccard index"},
    name="Jaccard Distance",
    description="Compute Jaccard distance for each sample "
                "in a feature table. Jaccard is calculated using"
                "presence/absence data. Data of type "
                "FeatureTable[Frequency | Relative Frequency] is reduced"
                "to presence/absence prior to calculation.",
    examples={"run on one core (by default)": examples.jaccard_example,
              "to run on n cores, replace 1 here with your preferred integer":
              examples.jaccard_n_jobs_example,
              "use 'auto' to run on all of host system's available CPU cores":
              examples.jaccard_auto_jobs_example,
              },
    citations=[citations['jaccard1908nouvelles']])


plugin.methods.register_function(
    function=beta.unweighted_unifrac,
    inputs={'table': FeatureTable[Frequency | RelativeFrequency
            | PresenceAbsence],
            'phylogeny': Phylogeny[Rooted]},
    parameters={'threads': Threads,
                'bypass_tips': Bool},
    outputs=[('distance_matrix', DistanceMatrix)],
    input_descriptions={
        'table': "The feature table containing the samples for which "
                 "Unweighted Unifrac should be computed.",
        'phylogeny': "Phylogenetic tree containing tip identifiers that "
                     "correspond to the feature identifiers in the table. "
                     "This tree can contain tip ids that are not present in "
                     "the table, but all feature ids in the table must be "
                     "present in this tree."},
    parameter_descriptions={
        'threads': threads_description,
        'bypass_tips':
            ("In a bifurcating tree, the tips make up about 50% of "
             "the nodes in a tree. By ignoring them, specificity "
             "can be traded for reduced compute time. This has the"
             " effect of collapsing the phylogeny, and is analogous"
             " (in concept) to moving from 99% to 97% OTUs")},
    output_descriptions={'distance_matrix': "Distance matrix for Unweighted "
                         "Unifrac."},
    name="Unweighted Unifrac",
    description="Compute Unweighted Unifrac for each sample in a "
                "feature table",
    examples={"run on one core (by default)": examples.u_u_example,
              "to run on n cores, replace 1 here with your preferred integer":
              examples.u_u_n_threads_example,
              "use 'auto' to run on all of host system's available CPU cores":
              examples.u_u_auto_threads_example,
              "use bypass_tips to trade specificity for reduced compute time":
              examples.u_u_bypass_tips_example,
              },
    citations=[
        citations['lozupone2005unifrac'],
        citations['lozupone2007unifrac'],
        citations['hamady2010unifrac'],
        citations['lozupone2011unifrac'],
        citations['mcdonald2018unifrac']]
)

plugin.methods.register_function(
    function=beta.weighted_unifrac,
    inputs={'table': FeatureTable[Frequency | RelativeFrequency],
            'phylogeny': Phylogeny[Rooted]},
    parameters={'threads': Threads,
                'bypass_tips': Bool},
    outputs=[('distance_matrix', DistanceMatrix)],
    input_descriptions={
        'table': "The feature table containing the samples for which "
                 "Weighted Unifrac should be computed.",
        'phylogeny': "Phylogenetic tree containing tip identifiers that "
                     "correspond to the feature identifiers in the table. "
                     "This tree can contain tip ids that are not present in "
                     "the table, but all feature ids in the table must be "
                     "present in this tree."},
    parameter_descriptions={
        'threads': threads_description,
        'bypass_tips':
            ("In a bifurcating tree, the tips make up about 50% of "
             "the nodes in a tree. By ignoring them, specificity "
             "can be traded for reduced compute time. This has the"
             " effect of collapsing the phylogeny, and is analogous"
             " (in concept) to moving from 99% to 97% OTUs")},
    output_descriptions={'distance_matrix': "Distance matrix for Unweighted "
                         "Unifrac."},
    name="Weighted Unifrac",
    description="Compute Weighted Unifrac for each sample in a "
                "feature table",
    examples={"run on one core (by default)": examples.w_u_example,
              "to run on n cores, replace 1 here with your preferred integer":
              examples.w_u_n_threads_example,
              "use 'auto' to run on all of host system's available CPU cores":
              examples.w_u_auto_threads_example,
              "use bypass_tips to trade specificity for reduced compute time":
              examples.w_u_bypass_tips_example,
              },
    citations=[
        citations['lozupone2005unifrac'],
        citations['lozupone2007unifrac'],
        citations['hamady2010unifrac'],
        citations['lozupone2011unifrac'],
        citations['mcdonald2018unifrac']]
)

# ------------------------ Passthrough Methods ------------------------
plugin.methods.register_function(
    function=alpha.alpha_passthrough,
    inputs={'table': FeatureTable[Frequency]},
    parameters={'metric': Str % Choices(alpha.METRICS['NONPHYLO']['UNIMPL'])},
    outputs=[('vector', SampleData[AlphaDiversity])],
    input_descriptions={'table': "The feature table containing the samples "
                        "for which a selected metric should be computed."},
    parameter_descriptions={'metric':
                            'The alpha diversity metric to be computed.'},
    output_descriptions={'vector': "Vector containing per-sample values "
                                   "for the chosen metric."},
    name="Alpha Passthrough (non-phylogenetic)",
    description="Computes a vector of values (one value for each samples in a "
                "feature table) using the scikit-bio implementation of the "
                "selected alpha diversity metric.",
    examples={'basic': examples.alpha_passthrough_example},
)


plugin.methods.register_function(
    function=beta.beta_passthrough,
    inputs={'table': FeatureTable[Frequency]},
    parameters={'metric': Str % Choices(beta.METRICS['NONPHYLO']['UNIMPL']),
                'pseudocount': Int % Range(1, None), 'n_jobs': Threads},
    outputs=[('distance_matrix', DistanceMatrix)],
    input_descriptions={
        'table': 'The feature table containing the samples over which beta '
                 'diversity should be computed.'
    },
    parameter_descriptions={
        'metric': 'The beta diversity metric to be computed.',
        'pseudocount': 'A pseudocount to handle zeros for compositional '
                       'metrics. This is ignored for non-compositional '
                       'metrics.',
        'n_jobs': n_jobs_description
    },
    output_descriptions={'distance_matrix': 'The resulting distance matrix.'},
    name='Beta Passthrough (non-phylogenetic)',
    description="Computes a distance matrix for all pairs of samples in a "
                "feature table using the scikit-bio implementation of the "
                "selected beta diversity metric.",
    examples={"run on one core (by default)":
              examples.beta_passthrough_example,
              "to run on n cores, replace 1 here with your preferred integer":
              examples.beta_passthrough_n_jobs_example,
              "use 'auto' to run on all of host system's available CPU cores":
              examples.beta_passthrough_auto_jobs_example,
              "use 'pseudocount' to manually set a pseudocount for " +
              "compositional metrics":
              examples.beta_passthrough_pseudocount_example,
              },
)


plugin.methods.register_function(
    function=beta.beta_phylogenetic_passthrough,
    inputs={'table': FeatureTable[Frequency],
            'phylogeny': Phylogeny[Rooted]},
    parameters={'metric': Str % Choices(beta.METRICS['PHYLO']['UNIMPL']),
                'threads': Threads,
                'variance_adjusted': Bool,
                'alpha': Float % Range(0, 1, inclusive_end=True),
                'bypass_tips': Bool},
    outputs=[('distance_matrix', DistanceMatrix)],
    input_descriptions={
        'table': 'The feature table containing the samples over which beta '
                 'diversity should be computed.',
        'phylogeny': 'Phylogenetic tree containing tip identifiers that '
                     'correspond to the feature identifiers in the table. '
                     'This tree can contain tip ids that are not present in '
                     'the table, but all feature ids in the table must be '
                     'present in this tree.'
    },
    parameter_descriptions={
        'metric': 'The beta diversity metric to be computed.',
        'threads': threads_description,
        'variance_adjusted': 'Perform variance adjustment based on Chang et '
                             'al. BMC Bioinformatics 2011. Weights distances '
                             'based on the proportion of the relative '
                             'abundance represented between the samples at a'
                             ' given node under evaluation.',
        'alpha': 'This parameter is only used when the choice of metric is '
                 'generalized_unifrac. The value of alpha controls importance'
                 ' of sample proportions. 1.0 is weighted normalized UniFrac.'
                 ' 0.0 is close to unweighted UniFrac, but only if the sample'
                 ' proportions are dichotomized.',
        'bypass_tips': 'In a bifurcating tree, the tips make up about 50% of '
                       'the nodes in a tree. By ignoring them, specificity '
                       'can be traded for reduced compute time. This has the '
                       'effect of collapsing the phylogeny, and is analogous '
                       '(in concept) to moving from 99% to 97% OTUs'
    },
    output_descriptions={'distance_matrix': 'The resulting distance matrix.'},
    name='Beta Phylogenetic Passthrough',
    description="Computes a distance matrix for all pairs of samples in a "
                "feature table using the unifrac implementation of the "
                "selected beta diversity metric.",
    examples={"run on one core (by default)":
              examples.beta_phylo_passthrough_example,
              "to run on n cores, replace 1 here with your preferred integer":
              examples.beta_phylo_passthrough_n_threads_example,
              "use 'auto' to run on all of host system's available CPU cores":
              examples.beta_phylo_passthrough_auto_threads_example,
              "use bypass_tips to trade specificity for reduced compute time":
              examples.beta_phylo_passthrough_bypass_tips_example,
              "variance adjustment":
              examples.beta_phylo_passthrough_variance_adjusted_example,
              "minimal generalized unifrac":
              examples.beta_phylo_passthrough_min_generalized_unifrac_example,
              "generalized unifrac":
              examples.beta_phylo_passthrough_generalized_unifrac_example,
              },
    citations=[
        citations['lozupone2005unifrac'],
        citations['lozupone2007unifrac'],
        citations['hamady2010unifrac'],
        citations['lozupone2011unifrac'],
        citations['mcdonald2018unifrac'],
        citations['chang2011variance'],
        citations['chen2012genUnifrac'],
        citations['sfiligoi2022unifrac']
    ]
)

plugin.methods.register_function(
    function=beta.beta_phylogenetic_meta_passthrough,
    inputs={'tables': List[FeatureTable[Frequency]],
            'phylogenies': List[Phylogeny[Rooted]]},
    parameters={'metric': Str % Choices(beta.METRICS['PHYLO']['UNIMPL']),
                'threads': Threads,
                'variance_adjusted': Bool,
                'alpha': Float % Range(0, 1, inclusive_end=True),
                'bypass_tips': Bool,
                'weights': List[Float],
                'consolidation': Str % Choices(list(CONSOLIDATIONS))},
    outputs=[('distance_matrix', DistanceMatrix)],
    input_descriptions={
        'tables': 'The feature tables containing the samples over which beta '
                  'diversity should be computed.',
        'phylogenies': 'Phylogenetic trees containing tip identifiers that '
                       'correspond to the feature identifiers in the table. '
                       'This tree can contain tip ids that are not present in '
                       'the table, but all feature ids in the table must be '
                       'present in this tree.'
    },
    parameter_descriptions={
        'metric': 'The beta diversity metric to be computed.',
        'threads': threads_description,
        'variance_adjusted': 'Perform variance adjustment based on Chang et '
                             'al. BMC Bioinformatics 2011. Weights distances '
                             'based on the proportion of the relative '
                             'abundance represented between the samples at a'
                             ' given node under evaluation.',
        'alpha': 'This parameter is only used when the choice of metric is '
                 'generalized_unifrac. The value of alpha controls importance'
                 ' of sample proportions. 1.0 is weighted normalized UniFrac.'
                 ' 0.0 is close to unweighted UniFrac, but only if the sample'
                 ' proportions are dichotomized.',
        'bypass_tips': 'In a bifurcating tree, the tips make up about 50% of '
                       'the nodes in a tree. By ignoring them, specificity '
                       'can be traded for reduced compute time. This has the '
                       'effect of collapsing the phylogeny, and is analogous '
                       '(in concept) to moving from 99% to 97% OTUs',
        'weights': 'The weight applied to each tree/table pair. This tuple is '
                   'expected to be in index order with tables and phylogenies.'
                   ' Default is to weight each tree/table pair evenly.',
        'consolidation': 'The matrix consolidation method, which determines '
                         'how the individual distance matrices are '
                         'aggregated'
    },
    output_descriptions={'distance_matrix': 'The resulting distance matrix.'},
    name='Beta Phylogenetic Meta Passthrough',
    description="Computes a distance matrix for all pairs of samples in the "
                "set of feature table and phylogeny pairs, using the unifrac "
                "implementation of the selected beta diversity metric.",
    examples={
        "Basic meta unifrac": examples.beta_phylo_meta_passthrough_example,
        "meta with weights": examples.beta_phylo_meta_weights_example,
        "changing the consolidation method":
        examples.beta_phylo_meta_consolidation_example,
    },
    citations=[
        citations['lozupone2005unifrac'],
        citations['lozupone2007unifrac'],
        citations['hamady2010unifrac'],
        citations['lozupone2011unifrac'],
        citations['mcdonald2018unifrac'],
        citations['chang2011variance'],
        citations['chen2012genUnifrac'],
        citations['lozupone2008metaunifrac']
    ]
)



# ----------------------------------------------------------------------------
# Copyright (c) 2018-2023, QIIME 2 development team.
#
# Distributed under the terms of the Modified BSD License.
#
# The full license is in the file LICENSE, distributed with this software.
# ----------------------------------------------------------------------------

from qiime2.plugin.testing import TestPluginBase


class UsageExampleTests(TestPluginBase):
    package = 'q2_diversity_lib.tests'

    def test_usage_examples(self):
        self.execute_examples()



# ----------------------------------------------------------------------------
# Copyright (c) 2018-2023, QIIME 2 development team.
#
# Distributed under the terms of the Modified BSD License.
#
# The full license is in the file LICENSE, distributed with this software.
# ----------------------------------------------------------------------------

from unittest import mock

import numpy as np
import numpy.testing as npt
import biom
import psutil
from qiime2 import Artifact
from qiime2.plugin.testing import TestPluginBase
from q2_types.feature_table import BIOMV210Format
from q2_types.tree import NewickFormat

from .._util import (_validate_requested_cpus,
                     _partition, _validate_tables)


class PartitionTests(TestPluginBase):
    package = 'q2_diversity_lib.tests'

    def test_table_below_batch_size(self):
        tab = biom.example_table.copy()
        partitions = list(_partition(tab, block_size=10))
        self.assertEqual(len(partitions), 1)
        self.assertEqual(tab, partitions[0])

    def test_table_above_batch_size(self):
        tab = biom.example_table.copy()
        partitions = list(_partition(tab, block_size=2))
        self.assertEqual(len(partitions), 2)
        npt.assert_equal(partitions[0].ids(), tab.ids()[:2])
        npt.assert_equal(partitions[1].ids(), tab.ids()[2])

    def test_table_equal_to_batch_size(self):
        tab = biom.example_table.copy()
        partitions = list(_partition(tab, block_size=3))
        self.assertEqual(len(partitions), 1)
        self.assertEqual(tab, partitions[0])


class ValidateTablesTests(TestPluginBase):
    package = 'q2_diversity_lib.tests'

    def setUp(self):
        super().setUp()
        self.empty_table = biom.Table(np.array([]), [], [])
        # empty table generated from self.empty_table with biom v2.1.7
        empty_table_fp = self.get_data_path('empty_table.biom')
        self.empty_table_as_BIOMV210Format = BIOMV210Format(empty_table_fp,
                                                            mode='r')
        valid_table_fp = self.get_data_path('crawford.biom')
        self.valid_table_as_BIOMV210Format = BIOMV210Format(valid_table_fp,
                                                            mode='r')
        not_a_table_fp = self.get_data_path('crawford.nwk')
        self.invalid_view_type = NewickFormat(not_a_table_fp, mode='r')

        self.valid_table_list = [self.valid_table_as_BIOMV210Format,
                                 self.valid_table_as_BIOMV210Format]
        self.invalid_table_list = [self.valid_table_as_BIOMV210Format,
                                   self.invalid_view_type]
        self.has_empty_table_list = [self.empty_table_as_BIOMV210Format,
                                     self.valid_table_as_BIOMV210Format]
        self.has_nan = biom.Table(np.array([[np.nan, 0, 1],
                                            [2, 3, 4]]),
                                  ['a', 'b'],
                                  ['x', 'y', 'z'])
        self.has_neg = biom.Table(np.array([[-1, 0, 1],
                                            [2, 3, 4]]),
                                  ['a', 'b'],
                                  ['x', 'y', 'z'])

        @_validate_tables
        def f1(table: biom.Table):
            pass
        self.function_with_table_param = f1

        @_validate_tables
        def f2():
            pass
        self.function_without_table_param = f2

    def test_pass_table_with_nan(self):
        with self.assertRaisesRegex(ValueError, "table.*contains NaN"):
            self.function_with_table_param(self.has_nan)

    def test_pass_table_with_negative_values(self):
        with self.assertRaisesRegex(ValueError,
                                    "table.*contains negative values"):
            self.function_with_table_param(self.has_neg)

    def test_pass_empty_table_positionally(self):
        with self.assertRaisesRegex(ValueError, "table.*is empty"):
            self.function_with_table_param(self.empty_table_as_BIOMV210Format)

    def test_pass_empty_table_positionally_list(self):
        with self.assertRaisesRegex(ValueError, "table.*is empty"):
            self.function_with_table_param(self.has_empty_table_list)

    def test_pass_empty_table_as_kwarg(self):
        with self.assertRaisesRegex(ValueError, "table.*is empty"):
            self.function_with_table_param(
                table=self.empty_table_as_BIOMV210Format)

    def test_pass_empty_table_as_kwarg_list(self):
        with self.assertRaisesRegex(ValueError, "table.*is empty"):
            self.function_with_table_param(
                table=self.has_empty_table_list)

    def test_decorated_lambda_with_table_param(self):
        with self.assertRaisesRegex(ValueError, "table.*is empty"):
            decorated_lambda = _validate_tables(lambda table: None)
            decorated_lambda(self.empty_table_as_BIOMV210Format)

    def test_decorated_lambda_with_table_param_list(self):
        with self.assertRaisesRegex(ValueError, "table.*is empty"):
            decorated_lambda = _validate_tables(lambda table: None)
            decorated_lambda(self.has_empty_table_list)

    def test_wrapped_function_has_no_table_param(self):
        with self.assertRaisesRegex(TypeError, "missing argument.*table"):
            self.function_without_table_param()

    def test_passed_invalid_view_type(self):
        with self.assertRaisesRegex(
                    ValueError, "Invalid view type.*Newick"):
            self.function_with_table_param(table=self.invalid_view_type)

    def test_passed_invalid_view_type_list(self):
        with self.assertRaisesRegex(
                    ValueError, "Invalid view type.*Newick"):
            self.function_with_table_param(table=self.invalid_table_list)


class ValidateRequestedCPUsTests(TestPluginBase):
    package = 'q2_diversity_lib.tests'

    def setUp(self):
        super().setUp()

        @_validate_requested_cpus
        def function_no_params():
            pass
        self.function_no_params = function_no_params

        @_validate_requested_cpus
        def function_w_param(n_jobs=3):
            return n_jobs
        self.function_w_n_jobs_param = function_w_param

        @_validate_requested_cpus
        def function_w_threads(threads=2):
            return threads
        self.function_w_threads_param = function_w_threads

        @_validate_requested_cpus
        def function_w_duplicate_params(n_jobs=3, threads=2):
            pass
        self.function_w_both = function_w_duplicate_params

        self.jaccard_thru_framework = self.plugin.actions['jaccard']
        self.unweighted_unifrac_thru_framework = self.plugin.actions[
                    'unweighted_unifrac']

        two_feature_table_fp = self.get_data_path('two_feature_table.biom')
        self.two_feature_table = biom.load_table(two_feature_table_fp)
        self.two_feature_table_as_BIOMV210Format = BIOMV210Format(
                two_feature_table_fp, mode='r')
        self.two_feature_table_as_artifact = Artifact.import_data(
                    'FeatureTable[Frequency]', two_feature_table_fp)

        larger_table_fp = self.get_data_path('crawford.biom')
        self.larger_table_as_artifact = Artifact.import_data(
                'FeatureTable[Frequency]', larger_table_fp)

        valid_tree_fp = self.get_data_path('three_feature.tree')
        self.valid_tree_as_NewickFormat = NewickFormat(valid_tree_fp, mode='r')
        self.valid_tree_as_artifact = Artifact.import_data(
                'Phylogeny[Rooted]', valid_tree_fp)

        larger_tree_fp = self.get_data_path('crawford.nwk')
        self.larger_tree_as_artifact = Artifact.import_data(
                'Phylogeny[Rooted]', larger_tree_fp)

    def test_function_without_cpu_request_param(self):
        with self.assertRaisesRegex(TypeError, 'without.*n_jobs.*threads'):
            self.function_no_params()

    @mock.patch("q2_diversity_lib._util.psutil.Process")
    def test_function_with_appropriate_param(self, mock_process):
        mock_process = psutil.Process()
        mock_process.cpu_affinity = mock.MagicMock(return_value=[0, 1, 2])
        self.assertEqual(self.function_w_n_jobs_param(3), 3)

        mock_process.cpu_affinity = mock.MagicMock(return_value=[4, 19])
        self.assertEqual(self.function_w_threads_param(2), 2)

    def test_function_with_duplicate_cpu_allocation_params(self):
        with self.assertRaisesRegex(TypeError, 'Duplicate parameters'):
            self.function_w_both()

    @mock.patch("q2_diversity_lib._util.psutil.Process")
    @mock.patch('psutil.cpu_count', return_value=999)
    def test_system_has_no_cpu_affinity(self, mock_cpu_count, mock_process):
        mock_process = psutil.Process()
        mock_process.cpu_affinity = mock.MagicMock(side_effect=AttributeError)
        self.assertEqual(self.function_w_n_jobs_param(999), 999)
        assert mock_process.cpu_affinity.called

    @mock.patch("q2_diversity_lib._util.psutil.Process")
    def test_requested_more_than_system_cpus(self, mock_process):
        mock_process = psutil.Process()
        mock_process.cpu_affinity = mock.MagicMock(return_value=[0, 1, 2])
        with self.assertRaisesRegex(ValueError, "\'n_jobs\' cannot exceed"):
            self.function_w_n_jobs_param(999)
        with self.assertRaisesRegex(ValueError, "\'threads\' cannot exceed"):
            self.function_w_threads_param(999)

    @mock.patch("q2_diversity_lib._util.psutil.Process")
    def test_requested_cpus_passed_as_kwarg(self, mock_process):
        mock_process = psutil.Process()
        mock_process.cpu_affinity = mock.MagicMock(return_value=[0, 1, 2])
        self.assertEqual(self.function_w_n_jobs_param(n_jobs=3), 3)

    @mock.patch("q2_diversity_lib._util.psutil.Process")
    def test_n_jobs_passed_as_default(self, mock_process):
        mock_process = psutil.Process()
        mock_process.cpu_affinity = mock.MagicMock(return_value=[0, 1, 2])
        self.assertEqual(self.function_w_n_jobs_param(), 3)

    @mock.patch("q2_diversity_lib._util.psutil.Process")
    def test_zero_passed_to_cpu_request(self, mock_process):
        mock_process = psutil.Process()
        mock_process.cpu_affinity = mock.MagicMock(return_value=[0, 1, 2])
        self.assertEqual(self.function_w_n_jobs_param(0), 3)
        self.assertEqual(self.function_w_n_jobs_param(n_jobs=0), 3)
        self.assertEqual(self.function_w_threads_param(0), 3)
        self.assertEqual(self.function_w_threads_param(threads=0), 3)

    @mock.patch("q2_diversity_lib._util.psutil.Process")
    def test_cpu_request_through_framework(self, mock_process):
        mock_process = psutil.Process()
        mock_process.cpu_affinity = mock.MagicMock(return_value=[0])

        self.jaccard_thru_framework(self.larger_table_as_artifact, n_jobs=1)
        self.jaccard_thru_framework(self.larger_table_as_artifact,
                                    n_jobs='auto')
        self.unweighted_unifrac_thru_framework(self.larger_table_as_artifact,
                                               self.larger_tree_as_artifact,
                                               threads=1)
        self.unweighted_unifrac_thru_framework(self.larger_table_as_artifact,
                                               self.larger_tree_as_artifact,
                                               threads='auto')
        # If we get here, then it ran without error
        self.assertTrue(True)

    @mock.patch("q2_diversity_lib._util.psutil.Process")
    def test_more_threads_than_max_stripes(self, mock_process):
        mock_process = psutil.Process()
        mock_process.cpu_affinity = mock.MagicMock(return_value=[0])

        # The two_feature_table used here has only three samples, meaning
        # that it has a max of (3+1)/2 = 2 stripes. Unifrac may report
        # requests of more-threads-than-stripes to stderror, but should handle
        # that situation gracefully.
        self.unweighted_unifrac_thru_framework(
                self.two_feature_table_as_artifact,
                self.valid_tree_as_artifact, threads=1)
        self.unweighted_unifrac_thru_framework(
                self.two_feature_table_as_artifact,
                self.valid_tree_as_artifact, threads='auto')
        # If we get here, then it ran without error
        self.assertTrue(True)



import numpy as np

from skbio.diversity._util import _validate_counts_vector
import skbio.diversity.alpha

from scipy.special import gammaln


# c&p methods from skbio
def _berger_parker(counts):
    counts = _validate_counts_vector(counts)
    return counts.max() / counts.sum()


def _brillouin_d(counts):
    counts = _validate_counts_vector(counts)
    nz = counts[counts.nonzero()]
    n = nz.sum()
    return (gammaln(n + 1) - gammaln(nz + 1).sum()) / n


def _simpsons_dominance(counts):
    counts = _validate_counts_vector(counts)
    return 1 - skbio.diversity.alpha.dominance(counts)


def _esty_ci(counts):
    counts = _validate_counts_vector(counts)

    f1 = skbio.diversity.alpha.singles(counts)
    f2 = skbio.diversity.alpha.doubles(counts)
    n = counts.sum()
    z = 1.959963985
    W = (f1 * (n - f1) + 2 * n * f2) / (n ** 3)

    return f1 / n - z * np.sqrt(W), f1 / n + z * np.sqrt(W)


def _goods_coverage(counts):
    counts = _validate_counts_vector(counts)
    f1 = skbio.diversity.alpha.singles(counts)
    N = counts.sum()
    return 1 - (f1 / N)


def _margalef(counts):
    counts = _validate_counts_vector(counts)
    # replaced observed_otu call to sobs
    return (skbio.diversity.alpha.sobs(counts) - 1) / np.log(counts.sum())


def _mcintosh_d(counts):
    counts = _validate_counts_vector(counts)
    u = np.sqrt((counts * counts).sum())
    n = counts.sum()
    return (n - u) / (n - np.sqrt(n))


def _strong(counts):
    counts = _validate_counts_vector(counts)
    n = counts.sum()
    # replaced observed_otu call to sobs
    s = skbio.diversity.alpha.sobs(counts)
    i = np.arange(1, len(counts) + 1)
    sorted_sum = np.sort(counts)[::-1].cumsum()
    return (sorted_sum / n - (i / s)).max()


def _p_evenness(counts):
    counts = _validate_counts_vector(counts)
    return _shannon(counts, base=np.e) / np.log(
        skbio.diversity.alpha.sobs(counts=counts))


def _shannon(counts, base=2):
    counts = _validate_counts_vector(counts)
    freqs = counts / counts.sum()
    nonzero_freqs = freqs[freqs.nonzero()]
    return -(nonzero_freqs * np.log(nonzero_freqs)).sum() / np.log(base)



# ----------------------------------------------------------------------------
# Copyright (c) 2018-2023, QIIME 2 development team.
#
# Distributed under the terms of the Modified BSD License.
#
# The full license is in the file LICENSE, distributed with this software.
# ----------------------------------------------------------------------------

from setuptools import setup, find_packages
import versioneer

setup(
    name="q2-diversity-lib",
    version=versioneer.get_version(),
    cmdclass=versioneer.get_cmdclass(),
    packages=find_packages(),
    author="Chris Keefe",
    author_email="crk239@nau.edu",
    description="Utility exposing diversity metrics/measures as Actions",
    entry_points={
        "qiime2.plugins":
        ["q2-diversity-lib=q2_diversity_lib.plugin_setup:plugin"]
    },
    url="https://qiime2.org",
    license="BSD-3-Clause",
    package_data={
        'q2_diversity_lib': ['citations.bib'],
        'q2_diversity_lib.tests': ['data/*']
    },
    zip_safe=False,
)



# ----------------------------------------------------------------------------
# Copyright (c) 2018-2023, QIIME 2 development team.
#
# Distributed under the terms of the Modified BSD License.
#
# The full license is in the file LICENSE, distributed with this software.
# ----------------------------------------------------------------------------

import biom
import numpy as np
import pkg_resources

from qiime2 import Artifact


def get_test_data_path(filename):
    return pkg_resources.resource_filename('q2_diversity_lib.tests',
                                           f'data/{filename}')


s_ids_1 = ['S1', 'S2', 'S3', 'S4', 'S5']
s_ids_2 = ['S6', 'S7', 'S8', 'S9', 'S10']


def ft1_factory():
    return Artifact.import_data(
        'FeatureTable[Frequency]',
        biom.Table(np.array([[1, 0, 5, 999, 1],
                            [0, 1, 2, 0, 5],
                            [0, 0, 0, 1, 10]]),
                   ['A', 'B', 'C'],
                   s_ids_1)
    )


def ft2_factory():
    return Artifact.import_data(
        'FeatureTable[Frequency]',
        biom.Table(np.array([[1, 10, 5, 999, 1],
                            [5, 1, 2, 0, 5],
                            [1, 40, 30, 10, 0]]),
                   ['A', 'B', 'C'],
                   s_ids_2)
    )


def tree_factory():
    input_tree_fp = get_test_data_path('faith_test.tree')
    return Artifact.import_data(
        'Phylogeny[Rooted]', input_tree_fp
    )


# ------------------------ alpha-diversity -----------------------
def faith_pd_example(use):
    ft = use.init_artifact('feature_table', ft1_factory)
    tree = use.init_artifact('phylogeny', tree_factory)
    result, = use.action(
        use.UsageAction(plugin_id='diversity_lib',
                        action_id='faith_pd'),
        use.UsageInputs(table=ft, phylogeny=tree),
        use.UsageOutputNames(vector='faith_pd_vector'))
    result.assert_output_type('SampleData[AlphaDiversity]')


def observed_features_example(use):
    ft = use.init_artifact('feature_table', ft1_factory)
    result, = use.action(
        use.UsageAction(plugin_id='diversity_lib',
                        action_id='observed_features'),
        use.UsageInputs(table=ft),
        use.UsageOutputNames(vector='obs_feat_vector'))
    result.assert_output_type('SampleData[AlphaDiversity]')

    # AssertRegex for lines in output files
    exp = zip(s_ids_1, [1, 1, 2, 2, 3])
    for id, val in exp:
        result.assert_has_line_matching(
            path='alpha-diversity.tsv',
            expression=f'{id}\t{val}'
        )


def pielou_evenness_example(use):
    ft = use.init_artifact('feature_table', ft1_factory)
    result, = use.action(
        use.UsageAction(plugin_id='diversity_lib',
                        action_id='pielou_evenness'),
        use.UsageInputs(table=ft),
        use.UsageOutputNames(vector='pielou_vector')
    )
    result.assert_output_type('SampleData[AlphaDiversity]')


def pielou_drop_example(use):
    ft = use.init_artifact('feature_table', ft1_factory)
    result, = use.action(
        use.UsageAction(plugin_id='diversity_lib',
                        action_id='pielou_evenness'),
        use.UsageInputs(table=ft, drop_undefined_samples=True),
        use.UsageOutputNames(vector='pielou_vector')
    )
    result.assert_output_type('SampleData[AlphaDiversity]')


def shannon_entropy_example(use):
    ft = use.init_artifact('feature_table', ft1_factory)
    result, = use.action(
        use.UsageAction(plugin_id='diversity_lib',
                        action_id='shannon_entropy'),
        use.UsageInputs(table=ft),
        use.UsageOutputNames(vector='shannon_vector')
    )
    result.assert_output_type('SampleData[AlphaDiversity]')


def shannon_drop_example(use):
    ft = use.init_artifact('feature_table', ft1_factory)
    result, = use.action(
        use.UsageAction(plugin_id='diversity_lib',
                        action_id='shannon_entropy'),
        use.UsageInputs(table=ft, drop_undefined_samples=True),
        use.UsageOutputNames(vector='shannon_vector')
    )
    result.assert_output_type('SampleData[AlphaDiversity]')


# ------------------------ beta-diversity -----------------------
def bray_curtis_example(use):
    ft = use.init_artifact('feature_table', ft1_factory)
    result, = use.action(
        use.UsageAction(plugin_id='diversity_lib',
                        action_id='bray_curtis'),
        use.UsageInputs(table=ft),
        use.UsageOutputNames(distance_matrix='bray_curtis_dm')
    )
    result.assert_output_type('DistanceMatrix')


def bray_curtis_n_jobs_example(use):
    ft = use.init_artifact('feature_table', ft1_factory)
    result, = use.action(
        use.UsageAction(plugin_id='diversity_lib',
                        action_id='bray_curtis'),
        use.UsageInputs(table=ft, n_jobs=1),
        use.UsageOutputNames(distance_matrix='bray_curtis_dm')
    )
    result.assert_output_type('DistanceMatrix')


def bray_curtis_auto_jobs_example(use):
    ft = use.init_artifact('feature_table', ft1_factory)
    result, = use.action(
        use.UsageAction(plugin_id='diversity_lib',
                        action_id='bray_curtis'),
        use.UsageInputs(table=ft, n_jobs='auto'),
        use.UsageOutputNames(distance_matrix='bray_curtis_dm')
    )
    result.assert_output_type('DistanceMatrix')


def jaccard_example(use):
    ft = use.init_artifact('feature_table', ft1_factory)
    result, = use.action(
        use.UsageAction(plugin_id='diversity_lib',
                        action_id='jaccard'),
        use.UsageInputs(table=ft),
        use.UsageOutputNames(distance_matrix='jaccard_dm')
    )
    result.assert_output_type('DistanceMatrix')


def jaccard_n_jobs_example(use):
    ft = use.init_artifact('feature_table', ft1_factory)
    result, = use.action(
        use.UsageAction(plugin_id='diversity_lib',
                        action_id='jaccard'),
        use.UsageInputs(table=ft, n_jobs=1),
        use.UsageOutputNames(distance_matrix='jaccard_dm')
    )
    result.assert_output_type('DistanceMatrix')


def jaccard_auto_jobs_example(use):
    ft = use.init_artifact('feature_table', ft1_factory)
    result, = use.action(
        use.UsageAction(plugin_id='diversity_lib',
                        action_id='jaccard'),
        use.UsageInputs(table=ft, n_jobs='auto'),
        use.UsageOutputNames(distance_matrix='jaccard_dm')
    )
    result.assert_output_type('DistanceMatrix')


def u_u_example(use):
    ft = use.init_artifact('feature_table', ft1_factory)
    tree = use.init_artifact('phylogeny', tree_factory)
    result, = use.action(
        use.UsageAction(plugin_id='diversity_lib',
                        action_id='unweighted_unifrac'),
        use.UsageInputs(table=ft, phylogeny=tree),
        use.UsageOutputNames(distance_matrix='unweighted_unifrac_dm')
    )
    result.assert_output_type('DistanceMatrix')


def u_u_n_threads_example(use):
    ft = use.init_artifact('feature_table', ft1_factory)
    tree = use.init_artifact('phylogeny', tree_factory)
    result, = use.action(
        use.UsageAction(plugin_id='diversity_lib',
                        action_id='unweighted_unifrac'),
        use.UsageInputs(table=ft, phylogeny=tree, threads=1),
        use.UsageOutputNames(distance_matrix='unweighted_unifrac_dm')
    )
    result.assert_output_type('DistanceMatrix')


def u_u_auto_threads_example(use):
    ft = use.init_artifact('feature_table', ft1_factory)
    tree = use.init_artifact('phylogeny', tree_factory)
    result, = use.action(
        use.UsageAction(plugin_id='diversity_lib',
                        action_id='unweighted_unifrac'),
        use.UsageInputs(table=ft, phylogeny=tree, threads='auto'),
        use.UsageOutputNames(distance_matrix='unweighted_unifrac_dm')
    )
    result.assert_output_type('DistanceMatrix')


def u_u_bypass_tips_example(use):
    ft = use.init_artifact('feature_table', ft1_factory)
    tree = use.init_artifact('phylogeny', tree_factory)
    use.comment("bypass_tips can be used with any threads setting, "
                "but auto may be a good choice if you're trimming run time.")
    result, = use.action(
        use.UsageAction(plugin_id='diversity_lib',
                        action_id='unweighted_unifrac'),
        use.UsageInputs(table=ft, phylogeny=tree,
                        threads='auto', bypass_tips=True),
        use.UsageOutputNames(distance_matrix='unweighted_unifrac_dm')
    )
    result.assert_output_type('DistanceMatrix')


def w_u_example(use):
    ft = use.init_artifact('feature_table', ft1_factory)
    tree = use.init_artifact('phylogeny', tree_factory)
    result, = use.action(
        use.UsageAction(plugin_id='diversity_lib',
                        action_id='weighted_unifrac'),
        use.UsageInputs(table=ft, phylogeny=tree),
        use.UsageOutputNames(distance_matrix='weighted_unifrac_dm')
    )
    result.assert_output_type('DistanceMatrix')


def w_u_n_threads_example(use):
    ft = use.init_artifact('feature_table', ft1_factory)
    tree = use.init_artifact('phylogeny', tree_factory)
    result, = use.action(
        use.UsageAction(plugin_id='diversity_lib',
                        action_id='weighted_unifrac'),
        use.UsageInputs(table=ft, phylogeny=tree, threads=1),
        use.UsageOutputNames(distance_matrix='weighted_unifrac_dm')
    )
    result.assert_output_type('DistanceMatrix')


def w_u_auto_threads_example(use):
    ft = use.init_artifact('feature_table', ft1_factory)
    tree = use.init_artifact('phylogeny', tree_factory)
    result, = use.action(
        use.UsageAction(plugin_id='diversity_lib',
                        action_id='weighted_unifrac'),
        use.UsageInputs(table=ft, phylogeny=tree, threads='auto'),
        use.UsageOutputNames(distance_matrix='weighted_unifrac_dm')
    )
    result.assert_output_type('DistanceMatrix')


def w_u_bypass_tips_example(use):
    ft = use.init_artifact('feature_table', ft1_factory)
    tree = use.init_artifact('phylogeny', tree_factory)
    use.comment("bypass_tips can be used with any threads setting, "
                "but auto may be a good choice if you're trimming run time.")
    result, = use.action(
        use.UsageAction(plugin_id='diversity_lib',
                        action_id='weighted_unifrac'),
        use.UsageInputs(table=ft, phylogeny=tree,
                        threads='auto', bypass_tips=True),
        use.UsageOutputNames(distance_matrix='weighted_unifrac_dm')
    )
    result.assert_output_type('DistanceMatrix')


# ------------------------ Passthrough Methods ------------------------
def alpha_passthrough_example(use):
    ft = use.init_artifact('feature_table', ft1_factory)
    result, = use.action(
        use.UsageAction(plugin_id='diversity_lib',
                        action_id='alpha_passthrough'),
        use.UsageInputs(table=ft, metric='simpson'),
        use.UsageOutputNames(vector='simpson_vector')
    )
    result.assert_output_type('SampleData[AlphaDiversity]')


def beta_passthrough_example(use):
    ft = use.init_artifact('feature_table', ft1_factory)
    result, = use.action(
        use.UsageAction(plugin_id='diversity_lib',
                        action_id='beta_passthrough'),
        use.UsageInputs(table=ft, metric='euclidean'),
        use.UsageOutputNames(distance_matrix='euclidean_dm')
    )
    result.assert_output_type('DistanceMatrix')


def beta_passthrough_n_jobs_example(use):
    ft = use.init_artifact('feature_table', ft1_factory)
    result, = use.action(
        use.UsageAction(plugin_id='diversity_lib',
                        action_id='beta_passthrough'),
        use.UsageInputs(table=ft, metric='euclidean', n_jobs=1),
        use.UsageOutputNames(distance_matrix='euclidean_dm')
    )
    result.assert_output_type('DistanceMatrix')


def beta_passthrough_auto_jobs_example(use):
    ft = use.init_artifact('feature_table', ft1_factory)
    use.comment("A default pseudocount of 1 is added to feature counts. "
                "Pseudocount is ignored for non-compositional metrics.")
    result, = use.action(
        use.UsageAction(plugin_id='diversity_lib',
                        action_id='beta_passthrough'),
        use.UsageInputs(table=ft, metric='aitchison', n_jobs='auto'),
        use.UsageOutputNames(distance_matrix='aitchison_dm')
    )
    result.assert_output_type('DistanceMatrix')


def beta_passthrough_pseudocount_example(use):
    ft = use.init_artifact('feature_table', ft1_factory)
    result, = use.action(
        use.UsageAction(plugin_id='diversity_lib',
                        action_id='beta_passthrough'),
        use.UsageInputs(table=ft, metric='aitchison', n_jobs='auto',
                        pseudocount=5),
        use.UsageOutputNames(distance_matrix='aitchison_dm')
    )
    result.assert_output_type('DistanceMatrix')


def beta_phylo_passthrough_example(use):
    ft = use.init_artifact('feature_table', ft1_factory)
    tree = use.init_artifact('phylogeny', tree_factory)
    result, = use.action(
        use.UsageAction(plugin_id='diversity_lib',
                        action_id='beta_phylogenetic_passthrough'),
        use.UsageInputs(table=ft, phylogeny=tree,
                        metric='weighted_normalized_unifrac'),
        use.UsageOutputNames(distance_matrix='weighted_normalized_unifrac_dm')
    )
    result.assert_output_type('DistanceMatrix')


def beta_phylo_passthrough_n_threads_example(use):
    ft = use.init_artifact('feature_table', ft1_factory)
    tree = use.init_artifact('phylogeny', tree_factory)
    result, = use.action(
        use.UsageAction(plugin_id='diversity_lib',
                        action_id='beta_phylogenetic_passthrough'),
        use.UsageInputs(table=ft, phylogeny=tree, threads=1,
                        metric='weighted_normalized_unifrac'),
        use.UsageOutputNames(distance_matrix='weighted_normalized_unifrac_dm')
    )
    result.assert_output_type('DistanceMatrix')


def beta_phylo_passthrough_auto_threads_example(use):
    ft = use.init_artifact('feature_table', ft1_factory)
    tree = use.init_artifact('phylogeny', tree_factory)
    result, = use.action(
        use.UsageAction(plugin_id='diversity_lib',
                        action_id='beta_phylogenetic_passthrough'),
        use.UsageInputs(table=ft, phylogeny=tree, threads='auto',
                        metric='weighted_normalized_unifrac'),
        use.UsageOutputNames(distance_matrix='weighted_normalized_unifrac_dm')
    )
    result.assert_output_type('DistanceMatrix')


def beta_phylo_passthrough_bypass_tips_example(use):
    ft = use.init_artifact('feature_table', ft1_factory)
    tree = use.init_artifact('phylogeny', tree_factory)
    use.comment("bypass_tips can be used with any threads setting, "
                "but auto may be a good choice if you're trimming run time.")
    result, = use.action(
        use.UsageAction(plugin_id='diversity_lib',
                        action_id='beta_phylogenetic_passthrough'),
        use.UsageInputs(table=ft, phylogeny=tree,
                        metric='weighted_normalized_unifrac',
                        threads='auto', bypass_tips=True),
        use.UsageOutputNames(distance_matrix='weighted_normalized_unifrac_dm')
    )
    result.assert_output_type('DistanceMatrix')


def beta_phylo_passthrough_variance_adjusted_example(use):
    use.comment(
        "Chang et al's variance adjustment may be applied to any unifrac "
        "method by using this passthrough function.")
    ft = use.init_artifact('feature_table', ft1_factory)
    tree = use.init_artifact('phylogeny', tree_factory)
    result, = use.action(
        use.UsageAction(plugin_id='diversity_lib',
                        action_id='beta_phylogenetic_passthrough'),
        use.UsageInputs(table=ft, phylogeny=tree,
                        metric='weighted_unifrac',
                        threads='auto',
                        variance_adjusted=True),
        use.UsageOutputNames(distance_matrix='var_adj_weighted_unifrac_dm')
    )
    result.assert_output_type('DistanceMatrix')


def beta_phylo_passthrough_min_generalized_unifrac_example(use):
    use.comment(
        "Generalized unifrac is passed alpha=1 by default. "
        "This is roughly equivalent to weighted normalized unifrac, "
        "which method will be used instead, because it is better optimized."
    )
    ft = use.init_artifact('feature_table', ft1_factory)
    tree = use.init_artifact('phylogeny', tree_factory)
    result, = use.action(
        use.UsageAction(plugin_id='diversity_lib',
                        action_id='beta_phylogenetic_passthrough'),
        use.UsageInputs(table=ft, phylogeny=tree,
                        metric='generalized_unifrac'),
        use.UsageOutputNames(distance_matrix='generalized_unifrac_dm')
    )
    result.assert_output_type('DistanceMatrix')


def beta_phylo_passthrough_generalized_unifrac_example(use):
    use.comment("passing a float between 0 and 1 to 'alpha' gives you control "
                "over the importance of sample proportions.")
    ft = use.init_artifact('feature_table', ft1_factory)
    tree = use.init_artifact('phylogeny', tree_factory)
    result, = use.action(
        use.UsageAction(plugin_id='diversity_lib',
                        action_id='beta_phylogenetic_passthrough'),
        use.UsageInputs(table=ft, phylogeny=tree,
                        metric='generalized_unifrac',
                        alpha=0.75),
        use.UsageOutputNames(distance_matrix='generalized_unifrac_dm')
    )
    result.assert_output_type('DistanceMatrix')


def beta_phylo_meta_passthrough_example(use):
    use.comment(
        "For brevity, these examples are focused on meta-specific parameters. "
        "See the documentation for beta_phylogenetic_passthrough for "
        "additional relevant information."
    )
    ft1 = use.init_artifact('feature_table1', ft1_factory)
    ft2 = use.init_artifact('feature_table2', ft2_factory)
    tree1 = use.init_artifact('phylogeny1', tree_factory)
    tree2 = use.init_artifact('phylogeny2', tree_factory)
    use.comment("NOTE: the number of trees and tables must match.")
    result, = use.action(
        use.UsageAction(plugin_id='diversity_lib',
                        action_id='beta_phylogenetic_meta_passthrough'),
        use.UsageInputs(tables=[ft1, ft2], phylogenies=[tree1, tree2],
                        metric='weighted_normalized_unifrac'),
        use.UsageOutputNames(distance_matrix='ft1_ft2_w_norm_unifrac_dm')
    )
    result.assert_output_type('DistanceMatrix')


def beta_phylo_meta_weights_example(use):
    ft1 = use.init_artifact('feature_table1', ft1_factory)
    ft2 = use.init_artifact('feature_table2', ft2_factory)
    tree = use.init_artifact('phylogeny', tree_factory)
    use.comment("The number of weights must match the number of tables/trees.")
    use.comment("If meaningful, it is possible to pass the same phylogeny "
                "more than once.")
    result, = use.action(
        use.UsageAction(plugin_id='diversity_lib',
                        action_id='beta_phylogenetic_meta_passthrough'),
        use.UsageInputs(tables=[ft1, ft2], phylogenies=[tree, tree],
                        metric='weighted_normalized_unifrac',
                        weights=[3.0, 42.0]),
        use.UsageOutputNames(distance_matrix='ft1_ft2_w_norm_unifrac_dm')
    )
    result.assert_output_type('DistanceMatrix')


def beta_phylo_meta_consolidation_example(use):
    ft1 = use.init_artifact('feature_table1', ft1_factory)
    ft2 = use.init_artifact('feature_table2', ft2_factory)
    tree1 = use.init_artifact('phylogeny1', tree_factory)
    tree2 = use.init_artifact('phylogeny2', tree_factory)
    result, = use.action(
        use.UsageAction(plugin_id='diversity_lib',
                        action_id='beta_phylogenetic_meta_passthrough'),
        use.UsageInputs(tables=[ft1, ft2], phylogenies=[tree1, tree2],
                        metric='weighted_normalized_unifrac',
                        weights=[0.4, 0.6],
                        consolidation='skipping_missing_values'),
        use.UsageOutputNames(distance_matrix='ft1_ft2_w_norm_unifrac_dm')
    )
    result.assert_output_type('DistanceMatrix')
